# -*- encoding: utf-8

"""
The main class that starts and runs wryten.
"""


import os
import sys

from . import ui
from . import module_loader
from . import themes
from . import helpers

from .file import File
from .logger import logger
from .config import Config
from .editor import Editor

__version__ = "1.42.10"


class App:
    def __init__(self, filenames=None, config_file=None):
        """
        Handle App initialization

        :param list filenames: Names of files to load initially
        :param str filenames[*]: Path to a file to load
        """
        self.version = __version__
        self.inited = False
        self.running = False
        self.debug = True
        self.block_rendering = False

        # Set default variables
        self.path = os.path.dirname(os.path.realpath(__file__))
        self.files = []
        self.current_file = 0
        self.status_msg = ""
        self.last_input = None
        self.global_buffer = []
        self.event_bindings = {}

        self.config = None
        self.ui = None
        self.modules = None
        self.themes = None

        # Maximum amount of inputs to process at once
        self.max_input = 100

        # Save filenames for later
        self.filenames = filenames

        # Save config file path for later
        self.config_file = config_file

        # Define core operations
        self.operations = {
            "help": self.help,
            "save_file": self.save_file,
            "run_command": self.query_command,
            "go_to": self.go_to,
            "open": self.open,
            "close_file": self.close_file,
            "new_file": self.new_file,
            "exit": self.ask_exit,
            "ask_exit": self.ask_exit,
            "prev_file": self.prev_file,
            "next_file": self.next_file,
            "save_file_as": self.save_file_as,
            "reload_file": self.reload_file,
            "toggle_mouse": self.toggle_mouse,
            "toggle_fullscreen": self.toggle_fullscreen,
        }

        # Bind our logger
        self.logger = logger
        self.logger.debug("Starting Suplemon...")

    def init(self):
        """Initialize the app."""
        # Load core components
        self.config = Config(self)
        if self.config_file:
            self.config.set_path(self.config_file)
        if not self.config.init():
            # Can't run without config
            return False
        self.config.load()

        # Unicode symbols don't play nice with Python 2 so disable them
        if sys.version_info[0] < 3:
            self.config["app"]["use_unicode_symbols"] = False

        # Configure logger
        self.debug = self.config["app"]["debug"]
        debug_level = self.config["app"]["debug_level"]
        self.logger.debug("Setting debug_level to {0}.".format(debug_level))
        self.logger.setLevel(debug_level)
        [handler.setLevel(debug_level) for handler in self.logger.handlers]

        # Load user interface
        self.ui = ui.UI(self)
        self.ui.init()

        # Load extension modules
        self.modules = module_loader.ModuleLoader(self)
        self.modules.load()

        # Load default module configs
        self.config.load_module_configs()

        # Load themes
        self.themes = themes.ThemeLoader(self)

        # Indicate that initialization is complete
        self.inited = True

        return True

    def exit(self):
        """Stop the main loop and exit."""
        self.trigger_event_before("app_exit")
        self.running = False

    def run(self):
        """Run the app via the ui wrapper."""
        self.ui.run(self.run_wrapped)

    def run_wrapped(self, *args):
        """Actually run the app and start mainloop.

        This shouldn't be called directly. Instead it's passed to the UI wich
        calls it in a safe curses wrapper.

        :param *args: Not used. Takes any args the wrapper might pass in.
        """
        # Load ui and files etc
        self.load()
        # Initial render
        self.get_editor().refresh()
        self.ui.refresh()
        # Start mainloop
        self.main_loop()
        self.trigger_event_after("app_exit")
        # Unload ui
        self.ui.unload()

    def load(self):
        """Load the app.

        Load the UI, open files in self.filenames and finally trigger
        the 'app_loaded' event.
        """
        self.ui.load()
        ver = sys.version_info
        if ver[0] < 3 or (ver[0] == 3 and ver[1] < 3):
            ver = ".".join(map(str, sys.version_info[0:2]))
            self.logger.warning("Running Suplemon with Python {version} "
                                "isn't officialy supported. Please use "
                                "Python 3.3 or higher."
                                .format(version=ver))
        self.load_files()
        self.running = True
        self.trigger_event_after("app_loaded")

    def on_input(self, event):
        # Handle the input or give it to the editor
        if not self.handle_input(event):
            # Pass the input to the editor component
            self.get_editor().handle_input(event)

    def main_loop(self):
        """Run the terminal IO loop until exit() is called."""
        while self.running:
            # Update ui before refreshing it
            self.ui.update()
            self.block_rendering = True
            got_input = False

            # Run through max 100 inputs (so the view is updated at least every 100 characters)
            for i in range(self.max_input):
                event = self.ui.get_input(False)  # non-blocking

                if not event:
                    break  # no more inputs to process at this time

                got_input = True
                self.on_input(event)

            if not got_input:
                # Wait for input, since there were none already available
                event = self.ui.get_input(True)  # blocking

                if event:
                    got_input = True
                    self.on_input(event)  # PERF: Up to 30% processing time

            self.block_rendering = False

            self.trigger_event_after("mainloop")
            # Rendering happens here
            # TODO: Optimize performance. Can make up 45% of processing time in the loop.
            self.get_editor().refresh()
            self.ui.refresh()

    def get_status(self):
        """Get the current status message.

        :return: Current status message.
        :rtype: str
        """
        return self.status_msg

    def get_file_index(self, file_obj):
        """Return the index of file_obj in the file list.

        :param file_obj: File instance.
        :return: Index of file_obj.
        :rtype: int
        """
        return self.files.index(file_obj)

    def get_key_bindings(self):
        """Return the list of key bindings."""
        return self.config.key_bindings

    def get_event_bindings(self):
        """Return the dict of event bindings."""
        return self.event_bindings

    def set_key_binding(self, key, operation):
        """Bind a key to an operation.

        Bind operation to be run when key is pressed.

        :param key: What key or key combination to bind.
        :param str operation: Which operation to run.
        """
        self.config.keymap.prepend({"keys": [key], "command": operation})

    def set_event_binding(self, event, when, callback):
        """Bind a callback to be run before or after an event.

        Bind callback to run before or after event occurs. The when parameter
        should be 'before' or 'after'. If using 'before' the callback can
        inhibit running the event if it returns True

        :param str event: Event to bind to.
        :param str when: String with 'before' or 'after'.
        :param callback: Callback to bind.
        """
        event_bindings = self.get_event_bindings()
        if when not in event_bindings.keys():
            event_bindings[when] = {}
        if event in event_bindings[when].keys():
            event_bindings[when][event].append(callback)
        else:
            event_bindings[when][event] = [callback]

    def set_status(self, status):
        """Set app status message.

        :param str status: Status message to show in status bar.
        """
        self.status_msg = str(status)

    def unsaved_changes(self):
        """Return True if there are unsaved changes in any file."""
        for f in self.files:
            if f.is_changed():
                return True
        return False

    def reload_config(self):
        """Reload configuration."""
        self.config.reload()
        for f in self.files:
            self.setup_editor(f.editor)
        self.trigger_event_after("config_loaded")
        self.ui.resize()
        self.ui.refresh()

    def handle_input(self, event):
        """Handle an input event.

        Runs relevant actions based on the event received.

        :param event: An event instance.
        :return: Boolean indicating if the event was handled.
        :rtype: boolean
        """
        if not event:
            return False
        self.last_input = event
        if event.type == "key":
            return self.handle_key(event)
        elif event.type == "mouse":
            return self.handle_mouse(event)
        return False

    def handle_key(self, event):
        """Handle a key input event.

        :param event: Event instance.
        :return: Boolean indicating if event was handled.
        :rtype: boolean
        """
        key_bindings = self.get_key_bindings()

        operation = None
        if event.key_name in key_bindings.keys():
            operation = key_bindings[event.key_name]
        elif event.key_code in key_bindings.keys():
            operation = key_bindings[event.key_code]

        if operation in self.operations.keys():
            self.run_operation(operation)
            return True
        elif operation in self.modules.modules.keys():
            self.run_module(operation)

        return False

    def handle_mouse(self, event):
        """Handle a mouse input event.

        :param event: Event instance.
        :return: Boolean indicating if event was handled.
        :rtype: boolean
        """
        editor = self.get_editor()
        if event.mouse_code == 1:                    # Left mouse button release
            editor.set_single_cursor(event.mouse_pos)
        elif event.mouse_code == 4096:               # Right mouse button release
            editor.add_cursor(event.mouse_pos)
        elif event.mouse_code == 524288:             # Wheel up
            editor.jump_up()
        elif event.mouse_code == 134217728:          # Wheel down(and unfortunately left button drag)
            editor.jump_down()
        else:
            return False
        return True

    ###########################################################################
    # User Interactions
    ###########################################################################

    def help(self):
        """Toggle the help document.
        - If current document is help, close it.
        - Otherwise, if help is open, switch to it.
        - Otherwise, open a new file with help text.
        """
        if self.get_file().is_help:
            self.close_file()
        else:
            idx = next((i for i, f in enumerate(self.files) if f.is_help), -1)
            if idx == -1:
                f = self.default_file()
                from . import help
                f.set_data(help.help_text)
                f.is_help = True
                self.files.append(f)
                idx = self.last_file_index()
            self.switch_to_file(idx)

    def new_file(self, path=None):
        """Open a new empty file.

        Open a new file and optionally set it's path.

        :param str path: Optional. Path for file.
        """
        new_file = self.default_file()
        if path:
            new_file.set_path(path)
        self.files.append(new_file)
        self.current_file = self.last_file_index()
        return new_file

    def ask_exit(self):
        """Exit if no unsaved changes, else make sure the user really wants to exit."""
        if self.unsaved_changes():
            yes = self.ui.query_bool("Exit?")
            if yes:
                self.exit()
                return True
            return False
        self.exit()
        return True

    def switch_to_file(self, index):
        """Load a default file if no files specified."""
        self.current_file = index

    def next_file(self):
        """Switch to next file."""
        if len(self.files) < 2:
            return
        cur = self.current_file
        cur += 1
        if cur > len(self.files)-1:
            cur = 0
        self.switch_to_file(cur)

    def prev_file(self):
        """Switch to previous file."""
        if len(self.files) < 2:
            return
        cur = self.current_file
        cur -= 1
        if cur < 0:
            cur = len(self.files)-1
        self.switch_to_file(cur)

    def go_to(self):
        """Go to a line or a file (or a line in a specific file with 'name:lineno')."""
        input_str = self.ui.query("Go to:")
        lineno = None
        fname = None
        if input_str is False:
            return False
        if input_str.find(":") != -1:
            parts = input_str.split(":")
            fname = parts[0]
            lineno = parts[1]
            file_index = self.find_file(fname)
            if file_index != -1:
                self.switch_to_file(file_index)
                try:
                    input_str = int(lineno)
                    self.get_editor().go_to_pos(input_str)
                except ValueError:
                    pass
        else:
            try:
                line_no = int(input_str)
                self.get_editor().go_to_pos(line_no)
            except ValueError:
                file_index = self.find_file(input_str)
                if file_index != -1:
                    self.switch_to_file(file_index)

    def find_file(self, s):
        """Return index of file matching string."""
        # REFACTOR: Move to a helper function or implement in a module

        # Case insensitive matching
        s = s.lower()

        # First match files beginning with s
        for i, file in enumerate(self.files):
            if file.name.lower().startswith(s):
                return i

        # Then match any files that contain s
        for i, file in enumerate(self.files):
            if s in file.name.lower():
                return i

        return -1

    def run_command(self, data):
        """Run editor commands."""
        parts = data.split(" ")
        cmd = parts[0].lower()
        if cmd in self.operations.keys():
            return self.run_operation(cmd)

        args = " ".join(parts[1:])
        self.logger.debug("Looking for command '{0}'".format(cmd))
        if cmd in self.modules.modules.keys():
            self.logger.debug("Trying to run command '{0}'".format(cmd))
            self.get_editor().store_action_state(cmd)
            if not self.run_module(cmd, args):
                return False
        else:
            self.set_status("Command '{0}' not found.".format(cmd))
            return False
        return True

    def run_module(self, module_name, args=""):
        try:
            self.modules.modules[module_name].run(self, self.get_editor(), args)
            return True
        except:
            # Catch any error when running a module just incase
            self.set_status("Running command failed!")
            self.logger.exception("Running command failed!")
            return False

    def run_operation(self, operation):
        """Run an app core operation."""
        # Support arbitrary callables. TODO: deprecate
        if hasattr(operation, "__call__"):
            return operation()

        if operation in self.operations.keys():
            cancel = self.trigger_event_before(operation)
            if not cancel:
                result = self.operations[operation]()
            self.trigger_event_after(operation)
            return result
        elif operation in self.modules.modules.keys():
            cancel = self.trigger_event_before(operation)
            if not cancel:
                result = self.modules.modules[operation].run(self, self.get_editor(), "")
            self.trigger_event_after(operation)
            return result

        return False

    def trigger_event(self, event, when):
        """Triggers event and runs registered callbacks."""
        status = False
        bindings = self.get_event_bindings()
        if when not in bindings.keys():
            return False
        if event in bindings[when].keys():
            callbacks = bindings[when][event]
            for cb in callbacks:
                try:
                    val = cb(event)
                except:
                    # Catch all errors in callbacks just incase
                    self.logger.error("Failed running callback: {0}".format(cb), exc_info=True)
                    continue
                if val:
                    status = True
        return status

    def trigger_event_before(self, event):
        return self.trigger_event(event, "before")

    def trigger_event_after(self, event):
        return self.trigger_event(event, "after")

    def toggle_fullscreen(self):
        """Toggle full screen editor."""
        display = self.config["display"]
        show_indicators = not display["show_top_bar"]
        display["show_top_bar"] = show_indicators
        display["show_bottom_bar"] = show_indicators
        display["show_legend"] = show_indicators
        # Virtual curses windows need to be resized
        self.ui.resize()

    def toggle_mouse(self):
        """Toggle mouse support."""
        # Invert the boolean
        self.config["editor"]["use_mouse"] = not self.config["editor"]["use_mouse"]
        self.ui.setup_mouse()
        if self.config["editor"]["use_mouse"]:
            self.set_status("Mouse enabled")
        else:
            self.set_status("Mouse disabled")

    def query_command(self):
        """Run editor commands."""
        if sys.version_info[0] < 3:
            modules = self.modules.modules.iteritems()
        else:
            modules = self.modules.modules.items()

        # Get built in operations
        completions = [oper for oper in self.operations.keys()]
        # Add runnable modules
        completions += [name for name, m in modules if m.is_runnable()]

        data = self.ui.query_autocmp("Command:", completions=sorted(completions))
        if not data:
            return False
        self.run_command(data)

    ###########################################################################
    # Editor operations
    ###########################################################################

    def new_editor(self):
        """Create a new editor instance."""
        editor = Editor(self, self.ui.editor_win)
        self.setup_editor(editor)
        return editor

    def get_editor(self):
        """Return the current editor."""
        return self.files[self.current_file].editor

    def setup_editor(self, editor):
        """Setup an editor instance with configuration."""
        config = self.config["editor"]
        editor.set_config(config)
        editor.init()

    ###########################################################################
    # File operations
    ###########################################################################

    def open(self):
        """Ask for file name and try to open it."""
        input_name = self.ui.query_file("Open file:")
        if not input_name:
            return False
        name_row_col = helpers.get_filename_cursor_pos(input_name)
        name = name_row_col["name"]
        exists = self.file_is_open(name)
        if exists:
            self.switch_to_file(self.files.index(exists))
            return True

        if not self.open_file(**name_row_col):
            self.set_status("Failed to load '{0}'".format(name))
            return False
        self.switch_to_file(self.last_file_index())
        return True

    def close_file(self):
        """Close current file if user confirms action."""
        if self.get_file().is_changed():
            if not self.ui.query_bool("Close file?"):
                return False
        self.files.pop(self.current_file)
        if not len(self.files):
            self.new_file()
            return False
        if self.current_file == len(self.files):
            self.current_file -= 1

    def save_file(self, file=False, overwrite=False):
        """Save current file."""
        f = file or self.get_file()
        # Make sure the file has a name
        if not f.get_name():
            return self.save_file_as(f)
        # Warn if the file has changed on disk
        if not overwrite and f.is_changed_on_disk():
            if not self.ui.query_bool("The file was modified since you opened it, save anyway?"):
                return False
        # Save the file
        if f.save():
            self.set_status("Saved [{0}] '{1}'".format(helpers.curr_time_sec(), f.name))
            if f.path() == self.config.path() or f.path() == self.config.keymap_path():
                self.reload_config()
            return True
        self.set_status("Couldn't write to '{0}'".format(f.name))
        return False

    def save_file_as(self, file=False):
        """Save current file."""
        f = file or self.get_file()
        name_input = self.ui.query_file("Save as:", f.name)
        if not name_input:
            return False

        target_dir, name = helpers.parse_path(name_input)
        full_path = os.path.join(target_dir, name)

        if os.path.exists(full_path):
            if not self.ui.query_bool("A file or directory with that name already exists. Overwrite it?"):
                return False
        if target_dir and not os.path.exists(target_dir):
            if self.ui.query_bool("The path doesn't exist, do you want to create it?"):
                self.logger.debug("Creating missing folders in save path.")
                os.makedirs(target_dir)
            else:
                return False
        f.set_path(full_path)
        # We can just overwrite the file since the user already confirmed
        return self.save_file(f, overwrite=True)

    def reload_file(self):
        """Reload the current file."""
        if self.ui.query_bool("Reload '{0}'?".format(self.get_file().name)):
            if self.get_file().reload():
                return True
        return False

    def get_files(self):
        """Return list of open files."""
        return self.files

    def get_file(self):
        """Return the current file."""
        return self.files[self.current_file]

    def last_file_index(self):
        """Get index of last file."""
        cur = len(self.files)-1
        return cur

    def current_file_index(self):
        """Get index of current file."""
        return self.current_file

    def open_file(self, name=None, row=0, col=0):
        """Open a file."""
        file = File(self)
        file.set_path(name)
        file.set_editor(self.new_editor())
        if not file.load():
            return False
        file.get_editor().set_single_cursor((col, row))
        file.get_editor().scroll_to_line(row)
        self.files.append(file)
        return True

    def load_files(self):
        """Try to load all files specified in arguments."""
        if self.filenames:
            for item in self.filenames:
                name_row_col = helpers.get_filename_cursor_pos(item)
                name = name_row_col["name"]
                if os.path.isdir(name):
                    continue
                # Avoid opening duplicate files
                if self.file_is_open(name):
                    continue
                if not self.open_file(**name_row_col):
                    self.new_file(name)
        # If nothing was loaded
        if not self.files:
            self.load_default()

    def file_is_open(self, path):
        """Check if file is open. Returns a File object or False."""
        for file in self.files:
            if file.path() == os.path.abspath(path):
                return file
        return False

    def load_default(self):
        """Load a default file if no files specified."""
        file = self.default_file()
        self.files.append(file)

    def default_file(self):
        """Create the default file."""
        file = File(self)
        file.set_editor(self.new_editor())
        # Specify contents to avoid appearing as modified
        file.set_data("")
        # Set markdown as the default file type
        file.editor.set_file_extension("md")
        return file



# -*- encoding: utf-8

import subprocess

from suplemon.suplemon_module import Module

import pyperclip

class SystemClipboard(Module):
    """Integrates the system clipboard with suplemon."""

    def init(self):
        self.init_logging(__name__)
        self.bind_event_before("insert", self.insert)
        self.bind_event_after("copy", self.copy)
        self.bind_event_after("cut", self.copy)

    def copy(self, event):
        lines = self.app.get_editor().get_buffer()
        data = "\n".join([str(line) for line in lines])
        self.set_clipboard(data)

    def insert(self, event):
        data = self.get_clipboard()
        lines = data.split("\n")
        self.app.get_editor().set_buffer(lines)

    def get_clipboard(self):
        return pyperclip.paste()

    def set_clipboard(self, data):
        pyperclip.copy(data)


module = {
    "class": SystemClipboard,
    "name": "system_clipboard",
}



# -*- encoding: utf-8

from suplemon.suplemon_module import Module


class Strip(Module):
    """Trim whitespace from start and end of lines."""

    def run(self, app, editor, args):
        line_nums = editor.get_lines_with_cursors()
        for n in line_nums:
            line = editor.lines[n]
            line.set_data(line.data.strip())


module = {
    "class": Strip,
    "name": "strip",
}



# -*- encoding: utf-8

import socket

from suplemon.suplemon_module import Module


class Hostname(Module):
    """Shows the machine hostname in the bottom status bar."""

    def init(self):
        self.hostname = ""
        hostinfo = None
        try:
            hostinfo = socket.gethostbyaddr(socket.gethostname())
        except:
            self.logger.debug("Failed to get hostname.")
        if hostinfo:
            self.hostname = hostinfo[0]
            # Use shorter hostname if available
            if hostinfo[1]:
                self.hostname = hostinfo[1][0]

    def get_status(self):
        if self.hostname:
            return "host:{0}".format(self.hostname)
        return ""


module = {
    "class": Hostname,
    "name": "hostname",
    "status": "bottom",
}



# -*- encoding: utf-8
"""
Text viewer component subclassed by Editor.
"""

import os
import re
import sys
import curses
import logging
try:
    import importlib
except ImportError:
    importlib = False

from . import helpers

from .line import Line
from .cursor import Cursor
from .themes import scope_to_pair
import suplemon.linelight  # NOQA

try:
    import pygments.lexers
    from .lexer import Lexer
except ImportError:
    pygments = False


class BaseViewer:
    def __init__(self, app, window):
        """
        Handle Viewer initialization

        :param App app: The main App class of Suplemon
        :param Window window: The ui window to use for the viewer
        """
        self.app = app
        self.window = window
        self.logger = logging.getLogger(__name__)
        self.config = {}
        self.data = ""
        self.lines = [Line()]
        self.file_extension = ""

        # Map special extensions to generic ones for highlighting
        self.extension_map = {
            "scss": "css",
            "less": "css",
            "tmtheme": "xml",
            "ts": "js",
        }
        self.show_line_ends = True

        self.cursor_style = curses.A_UNDERLINE

        self.y_scroll = 0
        self.x_scroll = 0
        self.cursors = [Cursor()]

        # Copy/paste buffer
        self.buffer = []

        # Last search used in 'find'
        self.last_find = ""

        # Runnable methods
        self.operations = {
            "arrow_right": self.arrow_right,              # Arrow Right
            "arrow_left": self.arrow_left,                # Arrow Left
            "arrow_up": self.arrow_up,                    # Arrow Up
            "arrow_down": self.arrow_down,                # Arrow Down
            "jump_left": self.jump_left,                  # Ctrl + Left
            "jump_right": self.jump_right,                # Ctrl + Right
            "jump_up": self.jump_up,                      # Ctrl + Up
            "jump_down": self.jump_down,                  # Ctrl + Down
            "page_up": self.page_up,                      # Page Up
            "page_down": self.page_down,                  # Page Down
            "home": self.home,                            # Home
            "end": self.end,                              # End
            "find": self.find_query,                      # Ctrl + F
            "find_next": self.find_next,                  # Ctrl + D
            "find_all": self.find_all,                    # Ctrl + A
        }

        self.pygments_syntax = None  # Needs to be implemented in derived classes
        self.lexer = None  # Needs to be implemented in derived classes

    def init(self):
        pass

    def get_buffer(self):
        """Returns the current buffer.

        Returns the local buffer or the global buffer depending on config.
        """
        if self.config["use_global_buffer"]:
            return self.app.global_buffer
        else:
            return self.buffer

    def get_size(self):
        """Get editor size (x,y)."""
        y, x = self.window.getmaxyx()
        return (x, y)

    @property
    def scroll_pos(self):
        return self.y_scroll, self.x_scroll

    @scroll_pos.setter
    def scroll_pos(self, pos):
        self.y_scroll = pos[0]
        self.x_scroll = pos[1]

    def get_cursor(self):
        """Return the main cursor."""
        return self.cursors[0]

    def get_first_cursor(self):
        """Get the first (primary) cursor."""
        highest = None
        for cursor in self.cursors:
            if highest is None or cursor.y < highest.y:
                highest = cursor
        return highest

    def get_last_cursor(self):
        """Get the last cursor."""
        lowest = None
        for cursor in self.cursors:
            if lowest is None:
                lowest = cursor
            elif cursor.y > lowest.y:
                lowest = cursor
            elif cursor.y == lowest.y and cursor.x > lowest.x:
                lowest = cursor
        return lowest

    def get_cursors_on_line(self, line_no):
        """Return all cursors on a specific line."""
        cursors = []
        for cursor in self.cursors:
            if cursor.y == line_no:
                cursors.append(cursor)
        return cursors

    def get_line(self, n):
        """Return line at index n.

        :param n: Index of line to get.
        :return: The Line instance.
        """
        return self.lines[n]

    def get_lines_with_cursors(self):
        """Return all line indices that have cursors.

        :return: A list of line numbers that have cursors.
        :rtype: list
        """
        line_nums = []
        for cursor in self.cursors:
            if cursor.y not in line_nums:
                line_nums.append(cursor.y)
        line_nums.sort()
        return line_nums

    def get_data(self):
        """Get editor contents.

        :return: Editor contents.
        :rtype: str
        """
        str_lines = []
        for line in self.lines:
            if isinstance(line, str):
                str_lines.append(line)
            else:
                str_lines.append(line.get_data())
        data = str(self.config["end_of_line"].join(str_lines))
        return data

    def set_data(self, data):
        """Set editor data or contents.

        :param str data: Set the editor contents to data.
        """
        self.data = data
        self.lines = []
        lines = self.data.splitlines()
        # splitlines doesn't return the last line if it's empty so we add a line
        # if there we're no lines at all or if the data ends with a new line
        if not len(lines) or self.data.endswith(("\n", "\r\n", "\r")):
            lines.append(Line())
        for line in lines:
            self.lines.append(Line(line))

    def set_config(self, config):
        """Set the viewer configuration dict.

        :param dict config: Editor config dict with any supported fields. See config.py.
        """
        self.config = config
        self.set_cursor_style(self.config["cursor_style"])

    def set_cursor_style(self, cursor_style):
        """Set cursor style.

        :param str cursor_style: Cursor type, either 'underline' or 'reverse'.
        """
        if cursor_style == "underline":
            self.cursor_style = curses.A_UNDERLINE
        elif cursor_style == "reverse":
            self.cursor_style = curses.A_REVERSE
        else:
            return False
        return True

    def set_cursor(self, cursor):
        self.logger.warning("set_cursor is deprecated, use set_cursor_style instead.")
        return self.set_cursor_style(cursor)

    def set_cursors(self, cursors):
        """Return list of all cursors."""
        self.cursors = [Cursor(c) for c in cursors]

    def set_single_cursor(self, cursor):
        """Discard all cursors and place a new one."""
        self.cursors = [Cursor(cursor)]

    def setup_linelight(self):
        raise NotImplementedError("Needs to be implemented in derived classes")

    def setup_highlight(self):
        raise NotImplementedError("Needs to be implemented in derived classes")

    def set_file_extension(self, ext):
        """Set the file extension."""
        ext = ext.lower()
        if ext and ext != self.file_extension:
            self.file_extension = ext
            self.setup_linelight()
            if self.config["show_highlighting"]:
                self.setup_highlight()

    def add_cursor(self, cursor):
        """Add a new cursor. Accepts a x,y tuple or a Cursor instance."""
        self.cursors.append(Cursor(cursor))

    def max_line_length(self):
        """Get maximum line length that fits in the editor."""
        return self.get_size()[0]-self.line_offset()-1

    def line_offset(self):
        """Get the x coordinate of beginning of line."""
        if not self.config["show_line_nums"]:
            return 0
        return len(str(len(self.lines)))+1

    def toggle_line_nums(self):
        """Toggle display of line numbers."""
        self.config["show_line_nums"] = not self.config["show_line_nums"]
        self.render()

    def toggle_line_ends(self):
        """Toggle display of line ends."""
        self.show_line_ends = not self.show_line_ends
        self.render()

    def toggle_highlight(self):
        """Toggle syntax highlighting."""
        return False

    ###########################################################################
    # Curses
    ###########################################################################

    def move_win(self, yx):
        """Move the editor window to position yx."""
        # Must try & catch since mvwin might
        # crash with incorrect coordinates
        try:
            self.window.mvwin(yx[0], yx[1])
        except:
            self.logger.warning("Moving window failed!", exc_info=True)

    def refresh(self):
        """Refresh the editor curses window."""
        self.move_cursors()
        self.render()
        self.window.refresh()

    def resize(self, yx=None):
        """Resize the UI."""
        if not yx:
            yx = self.window.getmaxyx()
        self.window.resize(yx[0], yx[1])
        self.move_cursors()
        self.refresh()

    def render(self):
        """Render the editor curses window."""
        if self.app.block_rendering:
            return

        self.window.erase()
        max_y = self.get_size()[1]
        max_len = self.max_line_length()
        lnum_len = self.line_offset() - 1
        lnum_pad = ">" if self.config["line_nums_pad_space"] else "0"
        # Iterate through visible lines
        for i in range(max_y):
            x_offset = self.line_offset()
            lnum = i + self.y_scroll
            if lnum >= len(self.lines):  # Make sure we have a line to show
                break

            # Get line for current row
            line = self.lines[lnum]

            attribs = None
            if self.config["highlight_current_line"] and self._is_current_line(i):
                # Highlight current line by adding bold to background attribute set
                attribs = self.window.getbkgd()
                self.window.bkgdset(" ", attribs | curses.A_BOLD)

            if self.config["show_line_nums"]:
                curs_color = curses.color_pair(line.number_color)
                padded_num = "{:{}{}d} ".format(lnum + 1, lnum_pad, lnum_len)
                self.window.addstr(i, 0, padded_num, curs_color)

            pos = (x_offset, i)
            try:
                self.render_line_contents(line, pos, x_offset, max_len)
            except:
                self.logger.error("Failed rendering line #{0} @{1} DATA:'{2}'!".format(lnum+1, pos, line),
                                  exc_info=True)
            if attribs is not None:
                # Restore background attribute set
                self.window.bkgdset(" ", attribs)

        self.render_cursors()

    def _is_current_line(self, y):
        for cursor in self.cursors:
            if cursor.y - self.y_scroll == y:
                return True
        return False

    def render_line_contents(self, line, pos, x_offset, max_len):
        """Render the contents of a line to the screen

        Renders a line to the screen with the appropriate rendering method
        based on settings.

        :param line: Line instance to render.
        :param pos: Position (x, y) for beginning of line.
        :param x_offset: Offset from left edge of screen. Currently same as x position.
        :param max_len: Maximum amount of chars that will fit on screen.
        """
        show_highlighting = self.config["show_highlighting"]
        if pygments and show_highlighting and self.pygments_syntax and self.app.themes.current_theme:
            self.render_line_pygments(line, pos, x_offset, max_len)
        elif self.config["show_line_colors"]:
            self.render_line_linelight(line, pos, x_offset, max_len)
        else:
            self.render_line_normal(line, pos, x_offset, max_len)

    def render_line_pygments(self, line, pos, x_offset, max_len):
        """Render line with Pygments syntax highlighting."""
        x, y = pos
        line_data = line.get_data()
        # Lazily prepare and slice the line,
        # even though it affects highlighting.
        line_data = self._prepare_line_for_rendering(line_data,
                                                     max_len,
                                                     no_wspace=True)
        # TODO:
        # 1) The line should not be prepared for rendering like this
        #    because it can get sliced. Sliced lines won't always get
        #    completely highlighted (partial words). Syntax highlighting
        #    should be done first and then only render visible words.
        # 2) Additionaly highlighing should be done for all lines at once
        #    and tokens should be cached in line instances. That way we can
        #    support multi line comment syntax etc. It should also perform
        #    better, since we only need to re-highlight lines when they change.
        # TODO 2: Optimize lexer performance
        tokens = self.lexer.lex(line_data, self.pygments_syntax)
        first_token = True
        for token in tokens:
            if token[1] == "\n":
                break
            scope = token[0]
            text = self.replace_whitespace(token[1])
            if token[1].isspace() and not self.app.ui.limited_colors:
                pair = 9  # Default to gray text on normal background
                settings = self.app.themes.get_scope("global")
                if settings and settings.get("invisibles"):
                    fg = int(settings.get("invisibles") or -1)
                    bg = int(settings.get("background") or -1)
                    curses.init_pair(pair, fg, bg)
                curs_color = curses.color_pair(pair)
                # Only add tab indicators to the inital whitespace
                if first_token and self.config["show_tab_indicators"]:
                    text = self.add_tab_indicators(text)
                self.window.addstr(y, x_offset, text, curs_color)
            else:
                # Color with pygments
                settings = self.app.themes.get_scope(scope)
                pair = scope_to_pair.get(scope)
                if settings and pair is not None:
                    fg = int(settings.get("foreground") or -1)
                    bg = int(settings.get("background") or -1)
                    curses.init_pair(pair, fg, bg)
                    curs_color = curses.color_pair(pair)
                    self.window.addstr(y, x_offset, text, curs_color)
                else:
                    self.window.addstr(y, x_offset, text)
            if first_token:
                first_token = False
            x_offset += len(text)

    def get_line_color(self, line):
        raise NotImplementedError("Needs to be implemented in derived classes")

    def render_line_linelight(self, line, pos, x_offset, max_len):
        """Render line with naive line based highlighting."""
        y = pos[1]
        line_data = line.get_data()
        line_data = self._prepare_line_for_rendering(line_data, max_len)
        curs_color = curses.color_pair(self.get_line_color(line))
        self.window.addstr(y, x_offset, line_data, curs_color)

    def render_line_normal(self, line, pos, x_offset, max_len):
        """Render line without any highlighting."""
        y = pos[1]
        line_data = line.get_data()
        line_data = self._prepare_line_for_rendering(line_data, max_len)
        self.window.addstr(y, x_offset, line_data)

    def add_tab_indicators(self, data):
        new_data = ""
        i = 0
        for char in data:
            if i == 0:
                new_data += self.config["tab_indicator_character"]
            else:
                new_data += char
            i += 1
            if i > self.config["tab_width"]-1:
                i = 0
        return new_data

    def replace_whitespace(self, data):
        # TODO: Optimize performance
        """Replace unsafe whitespace with alternative safe characters

        Replace unsafe whitespace with normal space or visible replacement.
        For example tab characters make cursors go out of sync with line
        contents.
        """
        for key in self.config["white_space_map"]:
            char = " "
            if self.config["show_white_space"]:
                char = self.config["white_space_map"][key]
            data = data.replace(key, char)
        # Remove newlines, they cause curses errors
        data = data.replace("\n", "")
        return data

    def _prepare_line_for_rendering(self, line_data, max_len, no_wspace=False):
        if self.show_line_ends:
            line_data += self.config["line_end_char"]
        line_data = self._slice_line_for_rendering(line_data, max_len)
        if not no_wspace:
            line_data = self.replace_whitespace(line_data)

        # Use unicode support on Python 3.3 and higher
        if sys.version_info[0] == 3 and sys.version_info[1] > 2:
            line_data = line_data.encode("utf-8")
        return line_data

    def _slice_line_for_rendering(self, line, max_len):
        """Return sliced line data.

        Returns what's left of line data after scrolling it horizontally
        and removing excess characters from the end.

        :param line: Line to slice.
        :param max_len: Maximum length of line.
        :return: Sliced line.
        """
        line = line[min(self.x_scroll, len(line)):]
        if not line:
            return ""
        # Clamp line length to view width
        line = line[:min(len(line), max_len)]
        return line

    def render_cursors(self):
        """Render editor window cursors."""
        if self.app.block_rendering:
            return

        max_x, max_y = self.get_size()
        for cursor in self.cursors:
            x = cursor.x - self.x_scroll + self.line_offset()
            y = cursor.y - self.y_scroll
            if y < 0:
                continue
            if y >= max_y:
                break
            if x < self.line_offset():
                continue
            if x > max_x-1:
                continue
            self.window.chgat(y, cursor.x+self.line_offset()-self.x_scroll, 1, self.cursor_style)

    ###########################################################################
    # Scrolling
    ###########################################################################

    def scroll_up(self):
        """Scroll view up if neccesary."""
        cursor = self.get_first_cursor()
        if cursor.y - self.y_scroll < 0:
            # Scroll up
            self.y_scroll = cursor.y

    def scroll_down(self):
        """Scroll view up if neccesary."""
        cursor = self.get_last_cursor()
        size = self.get_size()
        if cursor.y - self.y_scroll >= size[1]:
            # Scroll down
            self.y_scroll = cursor.y - size[1]+1

    def scroll_to_line(self, line_no):
        """Center the viewport on line_no."""
        if line_no >= len(self.lines):
            line_no = len(self.lines)-1
        new_y = line_no - int(self.get_size()[1] / 2)
        if new_y < 0:
            new_y = 0
        self.y_scroll = new_y

    def move_y_scroll(self, delta):
        """Add delta the y scroll axis scroll"""
        self.y_scroll += delta

    ###########################################################################
    # Cursors
    ###########################################################################

    def move_cursors(self, delta=None):
        """Move all cursors with delta. To avoid refreshing the screen set noupdate to True."""
        for cursor in self.cursors:
            if delta:
                if delta[0] != 0 and cursor.x >= 0:
                    cursor.move_right(delta[0])
                if delta[1] != 0 and cursor.y >= 0:
                    cursor.move_down(delta[1])

            if cursor.x < 0:
                cursor.x = 0
            if cursor.y < 0:
                cursor.y = 0
            if cursor.y >= len(self.lines)-1:
                cursor.y = len(self.lines)-1
            if cursor.x >= len(self.lines[cursor.y]):
                cursor.x = len(self.lines[cursor.y])
            elif cursor.persistent_x != cursor.x:
                # Retain the 'desired' x coordinate
                cursor.x = min(cursor.persistent_x, len(self.lines[cursor.y]))

        cur = self.get_cursor()  # Main cursor
        size = self.get_size()
        offset = self.line_offset()
        # Check if we should scroll horizontally
        if cur.x - self.x_scroll+offset > size[0] - 1:
            # -1 to allow space for cursor at line end
            self.x_scroll = len(self.lines[cur.y]) - size[0]+offset+1
        if cur.x - self.x_scroll < 0:
            self.x_scroll -= abs(cur.x - self.x_scroll)  # FIXME
        if cur.x - self.x_scroll+offset < offset:
            self.x_scroll -= 1
        self.purge_cursors()

    def move_x_cursors(self, line, col, delta):
        """Move all cursors starting at line and col with delta on the x axis."""
        for cursor in self.cursors:
            if cursor.y == line:
                if cursor.x > col:
                    cursor.move_right(delta)

    def move_y_cursors(self, line, delta, exclude=None):
        """Move all cursors starting at line and col with delta on the y axis.
        Exclude a cursor by passing it via the exclude argument."""
        for cursor in self.cursors:
            if cursor == exclude:
                continue
            if cursor.y > line:
                cursor.move_down(delta)

    def cursor_exists(self, cursor):
        """Check if a given cursor exists."""
        return cursor.tuple() in [cur.tuple() for cur in self.cursors]

    def remove_cursor(self, cursor):
        """Remove a cursor object from the cursor list."""
        try:
            index = self.cursors.index(cursor)
        except:
            return False
        self.cursors.pop(index)
        return True

    def purge_cursors(self):
        """Remove duplicate cursors that have the same position."""
        new = []
        # This sucks: can't use "if .. in .." for different instances (?)
        # Use a reference list instead. FIXME: use a generator
        ref = []
        for cursor in self.cursors:
            if not cursor.tuple() in ref:
                ref.append(cursor.tuple())
                new.append(cursor)
        self.cursors = new

    def purge_line_cursors(self, line_no):
        """Remove all but first cursor on given line."""
        line_cursors = []
        for cursor in self.cursors:
            if cursor.y == line_no:
                line_cursors.append(cursor)
        if len(line_cursors) < 2:
            return False

        # Leave the first cursor out
        line_cursors.pop(0)
        # Remove the rest
        for line_cursors in cursor:
            self.remove_cursor(cursor)
        return True

    ###########################################################################
    # Input Handling
    ###########################################################################

    def get_key_bindings(self):
        """Get list of editor key bindings."""
        return self.app.get_key_bindings()

    def handle_input(self, event):
        """Handle input."""
        if event.type == "mouse":
            return False
        key = event.key_code
        name = event.key_name
        # Try match a key to a method and call it

        key_bindings = self.get_key_bindings()
        operation = None
        if key in key_bindings:
            operation = key_bindings[key]
        elif name in key_bindings:
            operation = key_bindings[name]
        if operation:
            self.run_operation(operation)
            return True
        return False

    def run_operation(self, operation):
        """Run an editor core operation."""
        if operation in self.operations:
            cancel = self.app.trigger_event_before(operation)
            if cancel:
                return False
            result = self.operations[operation]()
            self.app.trigger_event_after(operation)
            return result
        return False

    ###########################################################################
    # Operations
    ###########################################################################

    def arrow_right(self):
        """Move cursors right."""
        for cursor in self.cursors:
            line = self.lines[cursor.y]
            # If we are at the end of the line
            if cursor.x >= len(line) or len(line) == 0:
                # If there is another line, then move down
                if cursor.y != len(self.lines)-1:
                    cursor.move_down()
                    cursor.set_x(0)
            # Otherwise, move the cursor right
            else:
                cursor.move_right()
        self.move_cursors()
        self.scroll_down()

    def arrow_left(self):
        """Move cursors left."""
        for cursor in self.cursors:
            if cursor.y != 0 and cursor.x == 0:
                cursor.move_up()
                cursor.set_x(len(self.lines[cursor.y])+1)
        self.move_cursors((-1, 0))
        self.scroll_up()

    def arrow_up(self):
        """Move cursors up."""
        self.move_cursors((0, -1))
        self.scroll_up()

    def arrow_down(self):
        """Move cursors down."""
        self.move_cursors((0, 1))
        self.scroll_down()

    def home(self):
        """Move to start of line or text on that line."""
        for cursor in self.cursors:
            wspace = helpers.whitespace(self.lines[cursor.y])
            if cursor.x == wspace:
                cursor.set_x(0)
            else:
                cursor.set_x(wspace)
        self.move_cursors()

    def end(self):
        """Move to end of line."""
        for cursor in self.cursors:
            cursor.set_x(len(self.lines[cursor.y]))
        self.move_cursors()

    def page_up(self):
        """Move half a page up."""
        amount = int(self.get_size()[1]/2) * -1
        self.move_cursors((0, amount))
        self.scroll_up()

    def page_down(self):
        """Move half a page down."""
        amount = int(self.get_size()[1]/2)
        self.move_cursors((0, amount))
        self.scroll_down()

    def jump_left(self):
        """Jump one 'word' to the left."""
        chars = self.config["punctuation"]
        for cursor in self.cursors:
            line = self.lines[cursor.y]
            if cursor.x == 0:
                if cursor.y > 0:
                    # Jump to end of previous line
                    cursor.set_x(len(self.lines[cursor.y-1]))
                    cursor.move_up()
                continue
            if cursor.x <= len(line):
                cur_chr = line[cursor.x-1]
            else:
                cur_chr = line[cursor.x]
            while cursor.x > 0:
                next = cursor.x-2
                if next < 0:
                    next = 0
                if cur_chr == " ":
                    cursor.move_left()
                    if line[next] != " ":
                        break
                else:
                    cursor.move_left()
                    if line[next] in chars:
                        break
        self.move_cursors()

    def jump_right(self):
        """Jump one 'word' to the right."""
        chars = self.config["punctuation"]
        for cursor in self.cursors:
            line = self.lines[cursor.y]
            if cursor.x == len(line):
                if cursor.y < len(self.lines):
                    # Jump to start of next line
                    cursor.set_x(0)
                    cursor.move_down()
                continue
            cur_chr = line[cursor.x]
            while cursor.x < len(line):
                next = cursor.x+1
                if next == len(line):
                    next -= 1
                if cur_chr == " ":
                    cursor.move_right()
                    if line[next] != " ":
                        break
                else:
                    cursor.move_right()
                    if line[next] in chars:
                        break
        self.move_cursors()

    def jump_up(self):
        """Jump up 3 lines."""
        self.move_cursors((0, -3))
        self.scroll_up()

    def jump_down(self):
        """Jump down 3 lines."""
        self.move_cursors((0, 3))
        self.scroll_down()

    def find_query(self):
        """Find in file via user input."""
        what = self.app.ui.query("Find:", self.last_find)
        if what:
            self.find(what)

    def store_action_state(self, state):
        raise NotImplementedError("Needs to be implemented in derived classes")

    def find(self, what, findall=False):
        """Find what in data (from top to bottom). Adds a cursor when found."""
        # Sorry for this colossal function
        if not what:
            return
        last_cursor = self.get_last_cursor()
        y = last_cursor.y

        found = False
        new_cursors = []
        # Loop through all lines starting from the last cursor
        while y < len(self.lines):
            line = self.lines[y]
            x_offset = 0  # Which character to begin searching from
            if y == last_cursor.y:
                # On the current line begin from the last cursor x pos
                x_offset = last_cursor.x

            # Find all occurances of search string
            s = str(line[x_offset:])  # Data to search in
            if y < len(self.lines)-1:  # Make line breaks findable (add them to all except last line)
                s = s + "\n"
            pattern = re.escape(what)  # Default to non regex pattern
            if self.config["regex_find"]:
                try:  # Try to search with the actual regex
                    indices = [match.start() for match in re.finditer(what, s)]
                except:  # Revert to normal search
                    indices = [match.start() for match in re.finditer(pattern, s)]
            else:
                # Use normal search
                indices = [match.start() for match in re.finditer(pattern, s)]

            # Loop through the indices and add cursors if they don't exist yet
            for i in indices:
                new = Cursor(i+x_offset, y)
                if not self.cursor_exists(new):
                    found = True
                    if new not in new_cursors:  # Make sure we don't get duplicates
                        new_cursors.append(new)
                    if not findall:
                        break
                if new not in new_cursors:
                    new_cursors.append(new)
            if found and not findall:
                break
            y += 1

        if not new_cursors:
            self.app.set_status("Can't find '{0}'".format(what))
            return False
        else:
            # If we only have one cursor, and it's not
            # where the first occurance is, just remove it
            if len(self.cursors) == 1 and self.cursors[0].tuple() != new_cursors[0].tuple():
                self.cursors = []
        self.last_find = what   # Only store string if it's really found

        # Add the new cursors
        for cursor in new_cursors:
            if not self.cursor_exists(cursor):
                self.cursors.append(cursor)

        destination = self.get_last_cursor().y
        self.scroll_to_line(destination)
        self.store_action_state("find")  # Store undo point

    def find_next(self):
        """Find next occurance."""
        what = self.last_find
        if what == "":
            cursor = self.get_cursor()
            search = "^([\w\-]+)"
            line = self.lines[cursor.y][cursor.x:]
            matches = re.match(search, line)
            if matches:
                what = matches.group(0)
            else:
                if line:
                    what = line[0]
            # Escape the data if regex is enabled
            if self.config["regex_find"]:
                what = re.escape(what)
            self.last_find = what
        self.find(what)

    def find_all(self):
        """Find all occurances."""
        self.find(self.last_find, True)


class Viewer(BaseViewer):
    def __init__(self, app, window):
        BaseViewer.__init__(self, app, window)

        # Lexer for translating tokens to strings
        self.lexer = None
        # Built in syntax definition (for commenting etc.)
        self.syntax = None
        # Normal Pygments lexer
        self.pygments_syntax = None

        self.setup_linelight()

    def init(self):
        if self.config["show_highlighting"]:
            self.setup_highlight()

    def setup_linelight(self):
        """Setup line based highlighting."""
        ext = self.file_extension
        # Check if a file extension is redefined
        # Maps e.g. 'scss' to 'css'
        if ext in self.extension_map:
            ext = self.extension_map[ext]  # Use it
        curr_path = os.path.dirname(os.path.realpath(__file__))

        filename = ext + ".py"
        path = os.path.join(curr_path, "linelight", filename)
        module = False
        syntax_module_name = ".{0}".format(ext)
        if importlib and os.path.isfile(path):
            try:
                module = importlib.import_module(syntax_module_name, "suplemon.linelight")
            except:
                self.logger.error("Failed to load syntax file '{0}'!".format(path), exc_info=True)
        else:
            return False

        if not module or "Syntax" not in dir(module):
            self.logger.error("File doesn't match API!")
            return False
        self.syntax = module.Syntax()
        self.pygments_syntax = None

    def setup_highlight(self):
        """Setup Pygments based highlighting."""
        if not pygments:
            return False
        self.lexer = Lexer(self.app)
        ext = self.file_extension.lower()
        if not ext:
            return False
        # Don't use Pygments for diffs. The text mate themes that are used don't often support it properly.
        # It's also such a basic format that it's justified to fall back on line based highlighting.
        if ext == "diff":
            return False
        # Check if a file extension is redefined
        # Maps e.g. 'scss' to 'css'
        if ext in self.extension_map:
            ext = self.extension_map[ext]  # Use it
        try:
            self.pygments_syntax = pygments.lexers.get_lexer_by_name(ext)
            self.logger.debug("Loaded Pygments lexer '{0}'.".format(ext))
        except:
            self.logger.debug("Failed to load Pygments lexer '{0}'.".format(ext))
            return False
        if ext == "php":
            # Hack to highlight PHP even without <?php ?> tags
            self.pygments_syntax.options.update({"startinline": 1})
            self.pygments_syntax.startinline = 1

    def get_line_color(self, raw_line):
        """Return a color based on line contents.

        :param str raw_line: The line from which to get a color value.
        :return: A color value for given raw_data.
        :rtype: int
        """
        if self.syntax:
            color = self.syntax.get_color(raw_line)
            if color is not None:
                return color
        return 0



# -*- encoding: utf-8


import hashlib

from suplemon.suplemon_module import Module


class ApplicationState(Module):
    """
    Stores the state of open files when exiting the editor and restores when files are reopened.

    Cursor positions and scroll position are stored and restored.
    """

    def init(self):
        self.init_logging(__name__)
        self.bind_event_after("app_loaded", self.on_load)
        self.bind_event_before("app_exit", self.on_exit)

    def on_load(self, event):
        """Runs when suplemon is fully loaded."""
        self.restore_states()

    def on_exit(self, event):
        """Runs before suplemon is exits."""
        self.store_states()

    def get_file_states(self):
        """Get the state of currently opened files. Returns a dict with the file path as key and file state as value."""
        states = {}
        for file in self.app.get_files():
            states[file.get_path()] = self.get_file_state(file)
        return states

    def get_file_state(self, file):
        """Get the state of a single file."""
        editor = file.get_editor()
        state = {
            "cursors": [cursor.tuple() for cursor in editor.cursors],
            "scroll_pos": editor.scroll_pos,
            "hash": self.get_hash(editor),
        }
        return state

    def get_hash(self, editor):
        # We don't need cryptographic security so we just use md5
        h = hashlib.md5()
        for line in editor.lines:
            h.update(line.get_data().encode("utf-8"))
        return h.hexdigest()

    def set_file_state(self, file, state):
        """Set the state of a file."""
        cursor = file.editor.get_cursor()
        # Don't set the cursor pos if it's not the default 0,0
        if cursor.x or cursor.y:
            return
        file.editor.set_cursors(state["cursors"])
        file.editor.scroll_pos = state["scroll_pos"]

    def store_states(self):
        """Store the states of opened files."""
        states = self.get_file_states()
        for path in states.keys():
            self.storage[path] = states[path]
        self.storage.store()

    def restore_states(self):
        """Restore the states of files that are currently open."""
        for file in self.app.get_files():
            path = file.get_path()
            if path in self.storage.get_data().keys():
                state = self.storage[path]
                if "hash" not in state:
                    self.set_file_state(file, state)
                elif state["hash"] == self.get_hash(file.get_editor()):
                    self.set_file_state(file, state)


module = {
    "class": ApplicationState,
    "name": "application_state",
}




# Default color levels for the color cube
cubelevels = [0x00, 0x5f, 0x87, 0xaf, 0xd7, 0xff]
# Generate a list of midpoints of the above list
snaps = [(x+y)/2 for x, y in list(zip(cubelevels, [0]+cubelevels))[1:]]


def hex_to_rgb(value):
    value = value.lstrip("#")
    lv = len(value)
    return tuple(int(value[i:i + lv // 3], 16) for i in range(0, lv, lv // 3))


def hex_to_xterm(hex):
    """
    Converts hex color values to the nearest equivalent xterm-256 color.
    """
    rgb = hex_to_rgb(hex)
    r, g, b = rgb
    # Using list of snap points, convert RGB value to cube indexes
    r, g, b = map(lambda x: len(tuple(s for s in snaps if s < x)), (r, g, b))
    # Simple colorcube transform
    return r*36 + g*6 + b + 16



# -*- encoding: utf-8

from suplemon.suplemon_module import Module


class Eval(Module):
    """
    Evaluate a python expression and show the result in the status bar.

    If no expression is provided the current line(s) are evaluated and
    replaced with the evaluation result.
    """

    def run(self, app, editor, args):
        if not args:
            return self.evaluate_lines(editor)
        else:
            return self.evaluate_input(args)

    def evaluate_input(self, inp):
        try:
            value = eval(inp)
        except:
            self.app.set_status("Eval failed.")
            return False
        self.app.set_status("Result:{0}".format(value))
        return True

    def evaluate_lines(self, editor):
        line_nums = editor.get_lines_with_cursors()
        for num in line_nums:
            line = editor.get_line(num)
            try:
                value = eval(line.get_data())
            except:
                continue
            line.set_data(str(value))


module = {
    "class": Eval,
    "name": "eval",
}



# -*- encoding: utf-8

from suplemon.suplemon_module import Module


class SaveAll(Module):
    """Save all currently open files. Asks for confirmation."""

    def run(self, app, editor, args):
        if not self.app.ui.query_bool("Save all files?", False):
            return False
        for file in app.get_files():
            file.save()


module = {
    "class": SaveAll,
    "name": "save_all",
}



# -*- encoding: utf-8

import re

from suplemon import helpers
from suplemon.suplemon_module import Module


class AutoComplete(Module):
    """
    A simple autocompletion module.

    This adds autocomplete support for the tab key. It uses a word
    list scanned from all open files for completions. By default it suggests
    the shortest possible match. If there are no matches, the tab action is
    run normally.
    """

    def init(self):
        self.word_list = []
        self.bind_event("tab", self.auto_complete)
        self.bind_event_after("app_loaded", self.build_word_list)
        self.bind_event_after("save_file", self.build_word_list)
        self.bind_event_after("save_file_as", self.build_word_list)

    def get_separators(self):
        """Return list of word separators obtained from app config.

        :return: String with all separators.
        :rtype: str
        """
        separators = self.app.config["editor"]["punctuation"]
        # Support words with underscores
        separators = separators.replace("_", "")
        return separators

    def build_word_list(self, *args):
        """Build the word list based on contents of open files."""
        word_list = []
        for file in self.app.files:
            data = file.get_editor().get_data()
            words = helpers.multisplit(data, self.get_separators())
            for word in words:
                # Discard undesired whitespace
                word = word.strip()
                # Must be longer than 1 and not yet in word_list
                if len(word) > 1 and word not in word_list:
                    word_list.append(word)
        self.word_list = word_list
        return False

    def get_match(self, word):
        """Find a completable match for word.

        :param word: The partial word to complete
        :return: The completion to add to the partial word
        :rtype: str
        """
        if not word:
            return False
        # Build list of suitable matches
        candidates = []
        for candidate in self.word_list:
            if candidate.startswith(word) and len(candidate) > len(word):
                candidates.append(candidate)
        # Find the shortest match
        # TODO: implement cycling through matches
        shortest = ""
        for candidate in candidates:
            if not shortest:
                shortest = candidate
                continue
            if len(candidate) < len(shortest):
                shortest = candidate
        if shortest:
            return shortest[len(word):]
        return False

    def run(self, app, editor, args):
        """Run the autocompletion."""
        self.auto_complete()

    def auto_complete(self, event):
        """Attempt to autocomplete at each cursor position.

        This callback runs before the tab action and tries to autocomplete
        the current word. If a match is found the tab action is inhibited.

        :param event: The event object.
        :return: True if a match is found.
        """
        editor = self.app.get_editor()
        pattern = "|".join(map(re.escape, self.get_separators()))
        matched = False
        for cursor in editor.cursors:
            line = editor.lines[cursor.y][:cursor.x]
            words = re.split(pattern, line)
            last_word = words[-1]
            match = self.get_match(last_word)
            if match:
                matched = True
                editor.type_at_cursor(cursor, match)
        return matched


module = {
    "class": AutoComplete,
    "name": "autocomplete",
}



# -*- encoding: utf-8
"""
Curses user interface.
"""

import os
import sys
import platform
import logging
from wcwidth import wcswidth

from .prompt import Prompt, PromptBool, PromptFiltered, PromptFile, PromptAutocmp
from .key_mappings import key_map

# Curses can't be imported yet but we'll
# predefine it to avoid confusing flake8
curses = None


class InputEvent:
    """Represents a keyboard or mouse event."""
    def __init__(self):
        self.type = None  # 'key' or 'mouse'
        self.key_name = ""
        self.key_code = None
        self.is_typeable = False
        self.curses_key_name = None
        self.mouse_code = None
        self.mouse_pos = (0, 0)
        self.logger = logging.getLogger("{0}.InputEvent".format(__name__))

    def parse_key_code(self, key_code):
        """Parse a key code (or character) from curses."""
        self.type = "key"
        self.key_code = key_code
        self.key_name = self._key_name(key_code)
        self.curses_key_name = self._curses_key_name(key_code)

    def set_key_name(self, name):
        """Manually set the event key name."""
        self.type = "key"
        self.key_name = name

    def parse_mouse_state(self, state):
        """Parse curses mouse events."""
        self.type = "mouse"
        self.mouse_code = state[4]
        self.mouse_pos = (state[1], state[2])

    def _key_name(self, key_code):
        """Return a normalized key name for key_code."""
        if isinstance(key_code, int):
            if key_code in key_map.keys():
                return key_map[key_code]
        curs_key_name = self._curses_key_name(key_code)
        if curs_key_name:
            if curs_key_name in key_map.keys():
                return key_map[curs_key_name]
            self.is_typeable = True  # We'll assume the key is typeable if it's not found in the key map
            return curs_key_name
        else:
            char = None
            if key_code in key_map.keys():
                return key_map[key_code]

            if sys.version_info[0] >= 3:
                if isinstance(key_code, str):
                    self.is_typeable = True
                    return key_code

            try:
                char = chr(key_code)
            except:
                pass
            if char is not None:
                self.is_typeable = True
                return char
        return False

    def _curses_key_name(self, key):
        """Return the curses key name for keys received from get_wch (and getch)."""
        # Handle multibyte get_wch input in Python 3.3
        if isinstance(key, str):
            return str(curses.keyname(ord(key)).decode("utf-8"))
        # Fallback to try and handle Python < 3.3
        # Special keys can also be ints on Python > 3.3
        if isinstance(key, int):  # getch fallback
            try:  # Try to convert to a curses key name
                name = str(curses.keyname(key).decode("utf-8"))
                return name
            except:  # Otherwise try to convert to a character
                return False
        return False

    def __str__(self):
        parts = [
            str(self.type),
            str(self.key_name),
            str(self.key_code),
            str(self.mouse_code),
            str(self.mouse_pos)
        ]
        return " ".join(parts)


class UI:
    def __init__(self, app):
        self.app = app
        self.logger = logging.getLogger(__name__)
        self.limited_colors = True
        self.screen = None
        self.current_yx = None
        self.text_input = None
        self.header_win = None
        self.status_win = None
        self.editor_win = None
        self.legend_win = None

    def init(self):
        """Set ESC delay and then import curses."""
        global curses
        # Set ESC detection time
        os.environ["ESCDELAY"] = str(self.app.config["app"]["escdelay"])
        termenv = os.environ.get("TERM", "")
        if termenv.endswith("-256color") and self.app.config["app"].get("imitate_256color"):
            # Curses doesn't recognize 'screen-256color' or 'tmux-256color' as 256-color terminals.
            # These terminals all seem to be identical to xterm-256color, which is recognized.
            # Since this might have other consequences it's hidden behind a config flag.
            self.logger.debug("Changing terminal from '{0}' to 'xterm-256color'".format(termenv))
            os.environ["TERM"] = "xterm-256color"
        # Now import curses, otherwise ESCDELAY won't have any effect
        import curses
        self.logger.debug("Loaded curses {0}".format(curses.version.decode()))

        # Notify user if Pygments syntax highlighting isn't available
        try:
            import pygments  # noqa
        except:
            self.logger.info("Pygments not available, please install python3-pygments for proper syntax highlighting.")

    def run(self, func):
        """Run the application main function via the curses wrapper for safety."""
        curses.wrapper(func)

    def load(self, *args):
        """Setup curses."""
        # Log the terminal type
        termname = curses.termname().decode("utf-8")
        self.logger.debug("Loading UI for terminal: {0}".format(termname))

        self.screen = curses.initscr()
        self.setup_colors()

        curses.raw()
        curses.noecho()

        try:
            # Hide the default cursor
            # Might fail on vt100 terminal emulators
            curses.curs_set(0)
        except:
            self.logger.warning("curses.curs_set(0) failed!")

        if "get_wch" not in dir(self.screen):
            self.logger.warning("Using old curses! Some keys and special characters might not work.")

        self.screen.keypad(1)

        self.current_yx = self.screen.getmaxyx()  # For checking resize
        self.setup_mouse()
        self.setup_windows()

    def unload(self):
        """Unload curses."""
        curses.endwin()

    def setup_mouse(self):
        # Mouse support
        curses.mouseinterval(10)
        if self.app.config["editor"]["use_mouse"]:
            curses.mousemask(-1)  # All events
        else:
            curses.mousemask(0)  # All events

    def setup_colors(self):
        """Initialize color support and define colors."""
        curses.start_color()
        try:
            curses.use_default_colors()
        except:
            self.logger.warning("Failed to load curses default colors. You could try 'export TERM=xterm-256color'.")
            return False

        # Default foreground color (could also be set to curses.COLOR_WHITE)
        fg = -1
        # Default background color (could also be set to curses.COLOR_BLACK)
        bg = -1

        # This gets colors working in TTY's as well as terminal emulators
        # curses.init_pair(10, -1, -1) # Default (white on black)
        # Colors for xterm (not xterm-256color)
        # Dark Colors

        if platform.system() != 'Windows':
            # it raises an exception on windows, cf https://github.com/zephyrproject-rtos/windows-curses/issues/10
            curses.init_pair(0, curses.COLOR_BLACK, bg)      # 0 Black
        curses.init_pair(1, curses.COLOR_RED, bg)        # 1 Red
        curses.init_pair(2, curses.COLOR_GREEN, bg)      # 2 Green
        curses.init_pair(3, curses.COLOR_YELLOW, bg)     # 3 Yellow
        curses.init_pair(4, curses.COLOR_BLUE, bg)       # 4 Blue
        curses.init_pair(5, curses.COLOR_MAGENTA, bg)    # 5 Magenta
        curses.init_pair(6, curses.COLOR_CYAN, bg)       # 6 Cyan
        curses.init_pair(7, fg, bg)                      # 7 White on Black
        curses.init_pair(8, fg, curses.COLOR_BLACK)      # 8 White on Black (Line number color)

        # Set color for whitespace
        # Fails on default Ubuntu terminal with $TERM=xterm (max 8 colors)
        # TODO: Smarter implementation for custom colors
        try:
            curses.init_pair(9, 8, bg)                   # Gray (Whitespace color)
            self.limited_colors = False
        except:
            # Try to revert the color
            self.limited_colors = True
            try:
                curses.init_pair(9, fg, bg)              # Try to revert color if possible
            except:
                # Reverting failed
                self.logger.error("Failed to set and revert extra colors.")

        # Nicer shades of same colors (if supported)
        if curses.can_change_color():
            try:
                # TODO: Define RGB for these to avoid getting
                # different results in different terminals
                # xterm-256color chart http://www.calmar.ws/vim/256-xterm-24bit-rgb-color-chart.html
                if platform.system() != 'Windows':
                    # it raises an exception on windows, cf https://github.com/zephyrproject-rtos/windows-curses/issues/10
                    curses.init_pair(0, 242, bg)  # 0 Black
                curses.init_pair(1, 204, bg)  # 1 Red
                curses.init_pair(2, 119, bg)  # 2 Green
                curses.init_pair(3, 221, bg)  # 3 Yellow
                curses.init_pair(4, 69, bg)   # 4 Blue
                curses.init_pair(5, 171, bg)  # 5 Magenta
                curses.init_pair(6, 81, bg)   # 6 Cyan
                curses.init_pair(7, 15, bg)   # 7 White
                curses.init_pair(8, 8, curses.COLOR_BLACK)  # 8 Gray on Black (Line number color)
                curses.init_pair(9, 8, bg)   # 8 Gray (Whitespace color)
            except:
                self.logger.warning("Enhanced colors failed to load. You could try 'export TERM=xterm-256color'.")
                self.app.config["editor"]["theme"] = "8colors"
        else:
            self.logger.warning("Enhanced colors not supported. You could try 'export TERM=xterm-256color'.")
            self.app.config["editor"]["theme"] = "8colors"

        self.app.themes.use(self.app.config["editor"]["theme"])

    def setup_windows(self):
        """Initialize and layout windows."""
        # We are using curses.newwin instead of self.screen.subwin/derwin because
        # subwindows are getting a special treatment on resize. e.g., the legend
        # bar may automatically resize to one line if the window gets smaller.
        # Even after doing the layout by moving the legend window into its proper
        # place a call to resize() to restore the 2 line height will error out.
        # This does not happen with curses.newwin().
        #
        # https://anonscm.debian.org/cgit/collab-maint/ncurses.git/tree/ncurses/base/resizeterm.c#n274
        # https://anonscm.debian.org/cgit/collab-maint/ncurses.git/tree/ncurses/base/wresize.c#n87
        self.text_input = None

        offset_top = 0
        offset_bottom = 0
        y, x = self.screen.getmaxyx()
        config = self.app.config["display"]

        if config["show_top_bar"]:
            offset_top += 1
            if self.header_win is None:
                self.header_win = curses.newwin(1, x, 0, 0)
            elif self.header_win.getmaxyx()[1] != x:
                # Header bar don't ever need to move
                self.header_win.resize(1, x)

        if config["show_bottom_bar"]:
            offset_bottom += 1
            if self.status_win is None:
                self.status_win = curses.newwin(1, x, y - offset_bottom, 0)
            else:
                self.status_win.mvwin(y - offset_bottom, 0)
                if self.status_win.getmaxyx()[1] != x:
                    self.status_win.resize(1, x)

        if config["show_legend"]:
            offset_bottom += 2
            if self.legend_win is None:
                self.legend_win = curses.newwin(2, x, y - offset_bottom, 0)
            else:
                self.legend_win.mvwin(y - offset_bottom, 0)
                if self.legend_win.getmaxyx()[1] != x:
                    self.legend_win.resize(2, x)

        if self.editor_win is None:
            self.editor_win = curses.newwin(y - offset_top - offset_bottom, x, offset_top, 0)
        else:
            # Not sure why editor implements this on its own
            # and if it's a good thing or not.
            self.app.get_editor().resize((y - offset_top - offset_bottom, x))
            self.app.get_editor().move_win((offset_top, 0))
            # self.editor_win.mvwin(offset_top, 0)
            # self.editor_win.resize(y - offset_top - offset_bottom, x)

    def get_size(self):
        """Get terminal size."""
        y, x = self.screen.getmaxyx()
        return (x, y)

    def update(self):
        self.check_resize()

    def refresh(self):
        self.refresh_status()
        self.screen.refresh()

    def resize(self, yx=None):
        """Resize UI to yx."""
        if yx is None:
            yx = self.screen.getmaxyx()
        self.screen.erase()
        curses.resize_term(yx[0], yx[1])
        self.setup_windows()

    def check_resize(self):
        """Check if terminal has resized and resize if needed."""
        yx = self.screen.getmaxyx()
        if self.current_yx != yx:
            self.current_yx = yx
            self.resize(yx)

    def refresh_status(self):
        """Refresh status windows."""
        if self.app.config["display"]["show_top_bar"]:
            self.show_top_status()
        if self.app.config["display"]["show_legend"]:
            self.show_legend()
        if self.app.config["display"]["show_bottom_bar"]:
            self.show_bottom_status()

    def show_top_status(self):
        """Show top status row."""
        self.header_win.erase()
        size = self.get_size()
        display = self.app.config["display"]
        head_parts = []
        if display["show_app_name"]:
            name_str = "Wryten IDE v{0} -".format(self.app.version)
            if self.app.config["app"]["use_unicode_symbols"]:
                logo = "\U0001f34b"  # Fancy lemon
                name_str = " {0} {1}".format(logo, name_str)
            head_parts.append(name_str)

        # Add module statuses to the status bar in descending order
        module_keys = sorted(self.app.modules.modules.keys())
        for name in module_keys:
            module = self.app.modules.modules[name]
            if module.options["status"] == "top":
                status = module.get_status()
                if status:
                    head_parts.append(status)

        if display["show_file_list"]:
            head_parts.append(self.file_list_str())

        head = " ".join(head_parts)
        head = head + (" " * (size[0]-wcswidth(head)-1))
        head_width = wcswidth(head)
        if head_width > size[0]:
            head = head[:size[0]-head_width]
        try:
            if self.app.config["display"]["invert_status_bars"]:
                self.header_win.addstr(0, 0, head, curses.color_pair(0) | curses.A_REVERSE)
            else:
                self.header_win.addstr(0, 0, head, curses.color_pair(0))
        except curses.error:
            pass
        self.header_win.refresh()

    def file_list_str(self):
        """Return rotated file list beginning at current file as a string."""
        curr_file_index = self.app.current_file_index()
        files = self.app.get_files()
        file_list = files[curr_file_index:] + files[:curr_file_index]
        str_list = []
        no_write_symbol = ["!", "\u2715"][self.app.config["app"]["use_unicode_symbols"]]
        is_changed_symbol = ["*", "\u2732"][self.app.config["app"]["use_unicode_symbols"]]
        for f in file_list:
            prepend = [no_write_symbol, ""][f.is_writable()]
            append = ""
            if self.app.config["display"]["show_file_modified_indicator"]:
                append += ["", is_changed_symbol][f.is_changed()]
            fname = prepend + f.name + append
            if not str_list:
                str_list.append("[{0}]".format(fname))
            else:
                str_list.append(fname)
        return " ".join(str_list)

    def show_bottom_status(self):
        """Show bottom status line."""
        editor = self.app.get_editor()
        size = self.get_size()
        cur = editor.get_cursor()

        # Core status info
        status_str = "@{0},{1} cur:{2} buf:{3}".format(
            str(cur[0]),
            str(cur[1]),
            str(len(editor.cursors)),
            str(len(editor.get_buffer()))
        )

        # Add module statuses to the status bar
        module_str = ""
        for name in self.app.modules.modules.keys():
            module = self.app.modules.modules[name]
            if module.options["status"] == "bottom":
                module_str += " " + module.get_status()
        status_str = module_str + " " + status_str

        self.status_win.erase()
        status = self.app.get_status()
        extra = size[0] - len(status+status_str) - 1
        line = status+(" "*extra)+status_str

        if len(line) >= size[0]:
            line = line[:size[0]-1]

        if self.app.config["display"]["invert_status_bars"]:
            attrs = curses.color_pair(0) | curses.A_REVERSE
        else:
            attrs = curses.color_pair(0)

        # This thwarts a weird crash that happens when pasting a lot
        # of data that contains line breaks into the find dialog.
        # Should probably figure out why it happens, but it's not
        # due to line breaks in the data nor is the data too long.
        # Thanks curses!
        try:
            self.status_win.addstr(0, 0, line, attrs)
        except:
            self.logger.exception("Failed to show bottom status bar. Status line was: {0}".format(line))

        self.status_win.refresh()

    def show_legend(self):
        """Show keyboard legend."""
        # Only the most important commands are displayed in the legend
        legend_commands = [
            ("save_file", "Save"),
            ("save_file_as", "Save as"),
            ("reload_file", "Reload"),
            ("undo", "Undo"),
            ("redo", "Redo"),
            ("open", "Open"),
            ("copy", "Copy"),
            ("cut", "Cut"),
            ("insert", "Paste"),
            ("find", "Find"),
            ("find_next", "Find next"),
            ("find_all", "Find all"),
            ("duplicate_line", "Duplicate line"),
            ("escape", "Single cursor"),
            ("go_to", "Go to"),
            ("run_command", "Run command"),
            ("toggle_mouse", "Mouse mode"),
            ("help", "Help"),
            ("ask_exit", "Exit"),
        ]

        # Get the key bindings for the commands
        keys = []
        for command in legend_commands:
            for item in self.app.config.keymap:
                if item["command"] == command[0]:
                    key = item["keys"][0]
                    keys.append((key, command[1]))
                    break

        # Render the keys
        self.legend_win.erase()
        x = 0
        y = 0
        max_y = 1
        for item in keys:
            key = item[0]
            label = item[1]
            key = key.upper()
            # Format some key names to look better
            if key.startswith("CTRL+"):
                key = "^"+key[5:]
            if key == "ESCAPE":
                key = "ESC"

            if x+len(" ".join((key, label))) >= self.get_size()[0]:
                x = 0
                y += 1
                if y > max_y:
                    break
            self.legend_win.addstr(y, x, key.upper(), curses.A_REVERSE)
            x += len(key)
            self.legend_win.addstr(y, x, " "+label)
            x += len(label)+2
        self.legend_win.refresh()

    def _query(self, text, initial="", cls=Prompt, inst=None):
        """Ask for text input via the status bar."""

        # Disable render blocking
        blocking = self.app.block_rendering
        self.app.block_rendering = 0

        # Create our text input
        if not inst:
            self.text_input = cls(self.app, self.status_win)
        else:
            self.text_input = inst
        self.text_input.set_config(self.app.config["editor"].copy())
        self.text_input.set_input_source(self.get_input)
        self.text_input.init()

        # Get input from the user
        out = self.text_input.get_input(text, initial)

        # Restore render blocking
        self.app.block_rendering = blocking

        return out

    def query(self, text, initial=""):
        """Get a single line input string from the user."""
        result = self._query(text, initial)
        return result

    def query_bool(self, text, default=False):
        """Get a boolean from the user."""
        result = self._query(text, default, PromptBool)
        return result

    def query_filtered(self, text, initial="", handler=None):
        """Get an arbitrary string from the user with input filtering."""
        prompt_inst = PromptFiltered(self.app, self.status_win, handler=handler)
        result = self._query(text, initial, inst=prompt_inst)
        return result

    def query_file(self, text, initial=""):
        """Get a file path from the user."""
        result = self._query(text, initial, PromptFile)
        return result

    def query_autocmp(self, text, initial="", completions=[]):
        """Get an arbitrary string from the user with autocomplete."""
        prompt_inst = PromptAutocmp(self.app, self.status_win, initial_items=completions)
        result = self._query(text, initial, inst=prompt_inst)
        return result

    def get_input(self, blocking=True):
        """Get an input event from keyboard or mouse. Returns an InputEvent instance or False."""
        event = InputEvent()  # Initialize new empty event
        char = False
        input_func = None
        if "get_wch" in dir(self.screen):
            # New Python 3.3 curses method for wide characters.
            input_func = self.screen.get_wch
        else:
            # Old Python fallback. No multibyte characters.
            input_func = self.screen.getch
        try:
            if blocking:
                self.screen.nodelay(0)
            else:
                self.screen.nodelay(1)
            char = input_func()
        except KeyboardInterrupt:
            # Handle KeyboardInterrupt as Ctrl+C
            event.set_key_name("ctrl+c")
            return event
        except:
            # No input available
            return False
        finally:
            self.screen.nodelay(0)

        if char and char != -1:
            if self.is_mouse(char):
                state = self.get_mouse_state()
                if state:
                    event.parse_mouse_state(state)
                    return event
            else:
                event.parse_key_code(char)
                return event
        return False

    def is_mouse(self, key):
        """Check for mouse events"""
        return key == curses.KEY_MOUSE

    def get_mouse_state(self):
        """Get the mouse event data."""
        try:
            mouse_state = curses.getmouse()
        except:
            self.logger.error("curses.getmouse() failed!", exc_info=True)
            return False

        # Translate the coordinates to the editor coordinate system
        return self._translate_mouse_to_editor(mouse_state)

    def _translate_mouse_to_editor(self, state):
        """Translate the screen coordinates to position in the editor view."""
        editor = self.app.get_editor()
        x, y = (state[1], state[2])
        if self.app.config["display"]["show_top_bar"]:
            y -= 1
        x -= editor.line_offset() - editor.x_scroll
        if x < 0:
            x = 0
        y += editor.y_scroll
        return (state[0], x, y, state[3], state[4])



from suplemon.linelight.color_map import color_map


class Syntax:
    def get_comment(self):
        return ("/*", "*/")

    def get_color(self, raw_line):
        color = color_map["white"]
        line = str(raw_line)
        if line.startswith("+"):
            color = color_map["green"]
        elif line.startswith("-"):
            color = color_map["red"]
        elif line.startswith("@@"):
            color = color_map["blue"]
        return color



# -*- encoding: utf-8

import pygments
import pygments.token
import pygments.lexers


class Lexer:
    def __init__(self, app):
        self.app = app
        self.token_map = {
            pygments.token.Comment: "comment",
            pygments.token.Comment.Single: "comment",
            pygments.token.Operator: "keyword",
            pygments.token.Name.Function: "entity.name.function",
            pygments.token.Name.Class: "entity.name.class",
            pygments.token.Name.Tag: "entity.name.tag",
            pygments.token.Name.Attribute: "entity.other.attribute-name",
            pygments.token.Name.Variable: "variable",
            pygments.token.Name.Builtin.Pseudo: "constant.language",
            pygments.token.Literal.String: "string",
            pygments.token.Literal.String.Doc: "string",
            pygments.token.Punctuation: "punctuation",
            pygments.token.Literal.Number: "constant.numeric",
            pygments.token.Name: "entity.name",
            pygments.token.Keyword: "keyword",
            pygments.token.Generic.Deleted: "invalid",
        }

    def lex(self, code, lex):
        """Return tokenified code.

        Return a list of tuples (scope, word) where word is the word to be
        printed and scope the scope name representing the context.

        :param str code: Code to tokenify.
        :param lex: Lexer to use.
        :return:
        """
        if lex is None:
            if not type(code) is str:
                # if not suitable lexer is found, return decoded code
                code = code.decode("utf-8")
            return (("global", code),)

        words = pygments.lex(code, lex)

        scopes = []
        for word in words:
            token = word[0]
            scope = "global"

            if token in self.token_map.keys():
                scope = self.token_map[token]

            scopes.append((scope, word[1]))
        return scopes



# -*- encoding: utf-8
"""
Line object to represent a single line in the text editor.
"""


class Line:
    def __init__(self, data=""):
        if isinstance(data, Line):
            data = data.data
        self.data = data
        self.x_scroll = 0
        self.number_color = 8

    def __getitem__(self, i):
        return self.data[i]

    def __setitem__(self, i, v):
        self.data[i] = v

    def __str__(self):
        return self.data

    def __add__(self, other):
        return Line(self.data + other)

    def __radd__(self, other):
        return Line(other + self.data)

    def __len__(self):
        return len(self.data)

    def get_data(self):
        return self.data

    def set_data(self, data):
        if isinstance(data, Line):
            data = data.get_data()
        self.data = data

    def set_number_color(self, color):
        self.number_color = color

    def find(self, what, start=0):
        return self.data.find(what, start)

    def strip(self, *args):
        return self.data.strip(*args)

    def reset_number_color(self):
        self.number_color = 8



#!/usr/bin/env python3
# -*- encoding: utf-8
"""
Start a Suplemon instance in the current window
"""

import sys

try:
    import argparse
except:
    # Python < 2.7
    argparse = False

from .main import App, __version__


def main():
    """Handle CLI invocation"""
    # Parse our CLI arguments
    config_file = None
    if argparse:
        parser = argparse.ArgumentParser(description="Console text editor with multi cursor support")
        parser.add_argument("filenames", metavar="filename", type=str, nargs="*", help="Files to load into Suplemon")
        parser.add_argument("--version", action="version", version=__version__)
        parser.add_argument("--config", type=str, help="Configuration file path.")
        args = parser.parse_args()
        filenames = args.filenames
        config_file = args.config
    else:
        # Python < 2.7 fallback
        filenames = sys.argv[1:]

    # Generate and start our application
    app = App(filenames=filenames, config_file=config_file)
    if app.init():
        app.run()

    # Output log info
    if app.debug:
        for logger_handler in app.logger.handlers:
            logger_handler.close()


if __name__ == "__main__":
    main()



from suplemon.linelight.color_map import color_map


class Syntax:
    def get_comment(self, line):
        return ("", "")

    def get_color(self, raw_line):
        color = color_map["white"]
        line = raw_line.strip()
        if line.startswith(("*", "-")):  # List
            color = color_map["cyan"]
        elif line.startswith("#"):  # Header
            color = color_map["green"]
        elif line.startswith(">"):  # Item description
            color = color_map["yellow"]
        elif raw_line.startswith("    "):  # Code
            color = color_map["magenta"]
        return color



# -*- encoding: utf-8

import time

from suplemon.suplemon_module import Module


class Clock(Module):
    """Shows a clock in the top status bar."""

    def get_status(self):
        s = time.strftime("%H:%M")
        if self.app.config["app"]["use_unicode_symbols"]:
            return "\u231A" + s
        return s


module = {
    "class": Clock,
    "name": "clock",
    "status": "top",
}



# -*- encoding: utf-8

from suplemon.suplemon_module import Module


class TabsToSpaces(Module):
    """Convert tab characters to spaces in the entire file."""

    def run(self, app, editor, args):
        for i, line in enumerate(editor.lines):
            new = line.data.replace("\t", " "*editor.config["tab_width"])
            editor.lines[i].set_data(new)


module = {
    "class": TabsToSpaces,
    "name": "tabstospaces",
}



# -*- encoding: utf-8

from suplemon.suplemon_module import Module


class SortLines(Module):
    """
    Sort current lines.

    Sorts alphabetically by default.
    Add 'length' to sort by length.
    Add 'reverse' to reverse the sorting.
    """

    def sort_normal(self, line):
        return line.data

    def sort_length(self, line):
        return len(line.data)

    def run(self, app, editor, args):
        args = args.lower()

        sorter = self.sort_normal
        reverse = True if "reverse" in args else False
        if "length" in args:
            sorter = self.sort_length

        indices = editor.get_lines_with_cursors()
        lines = [editor.get_line(i) for i in indices]

        sorted_lines = sorted(lines, key=sorter, reverse=reverse)

        for i, line in enumerate(sorted_lines):
            editor.lines[indices[i]] = line


module = {
    "class": SortLines,
    "name": "sort_lines",
}



# -*- encoding: utf-8

from suplemon import helpers
from suplemon.suplemon_module import Module


class AutoDocstring(Module):
    """
    Simple module for adding docstring placeholders.

    This module is intended to generate docstrings for Python functions.
    It adds placeholders for descriptions, arguments and return data.
    Function arguments are crudely parsed from the function definition
    and return statements are scanned from the function body.
    """

    def init(self):
        self.init_logging(__name__)
        self.default_values = {
            "short_desc": "Short description.",
            "long_desc": "Long description.",
            "args": "",
            "returns": "",
        }
        self.docstring_template = "{short_desc}\n" + \
            "\n" + \
            "{long_desc}\n" + \
            "\n" + \
            "{args}" + \
            "{returns}"

    def run(self, app, editor, args):
        """Run the autosphinx command."""
        cursor = editor.get_cursor()
        line = editor.get_line(cursor.y)
        line_data = line.get_data()
        if not line_data.strip().startswith("def "):
            app.set_status("Current line isn't a function definition.")
            return False

        func_args = self.get_function_args(line_data)
        func_returns = self.get_function_returns(editor, cursor.y)
        docstring = self.get_docstring(func_args=func_args, func_returns=func_returns)

        indent = self.get_docstring_indent(line_data)
        indent = indent * self.app.config["editor"]["tab_width"] * " "
        self.insert_docstring(editor, cursor.y+1, indent, docstring)

    def insert_docstring(self, editor, line_number, indent, docstring):
        """Insert a docstring at a specific line.

        Inserts the given docstring into editor at a specific
        line and indentation.

        :param editor: The editor to insert into
        :param int line_number: The first line number
        :param str indent: How much the docstring
        :param str docstring: The actual docstring
        """
        docstring = '"""{0}"""'.format(docstring)
        raw_lines = docstring.split("\n")
        lines = []
        for line in raw_lines:
            if not line.strip():
                lines.append("")
            else:
                lines.append(indent+line)
        editor.insert_lines_at(lines, line_number)

    def get_docstring_indent(self, line_data):
        """Get indent amount for docstring.

        Gets the indentation of the function definiton line, and
        adds +1 to account for the function body.

        :param line_data: The line of the function definition.
        """
        tab = self.app.config["editor"]["tab_width"]
        wspace = helpers.whitespace(line_data)
        indent = int(wspace/tab)+1
        return indent

    def get_docstring(self, **values):
        for key in self.default_values.keys():
            if key not in values.keys():
                values[key] = self.default_values[key]

        args = ""
        for arg in values["func_args"]:
            args += ":param {0}:\n".format(arg)
        values["args"] = args

        if values["func_returns"]:
            values["returns"] = ":return:\n"

        doc = self.docstring_template
        doc = doc.format(**values)
        return doc

    def get_function_name(self, line_data):
        """Get the name of function defined in line_data.

        :param str line_data: The line containing the function definition.
        """
        return helpers.get_string_between("def ", "(", line_data)

    def get_function_args(self, line_data):
        """Get list of arguments for function

        Parses a function definition in line_data and returns argument list.

        :param str line_data: The line containing the function definition.
        """
        func_name = self.get_function_name(line_data)
        raw_args = helpers.get_string_between(func_name+"(", "):", line_data)
        parts = raw_args.split(",")
        args = [part.strip() for part in parts]
        if "self" in args:
            args.pop(args.index("self"))
        return args

    def get_function_returns(self, editor, line_number):
        """Returns True if function at line_number returns something.

        :param editor: Editor instance to get lines from.
        :param line_number: Line number of the function definition.
        :return: Boolean indicating wether the function something.
        """
        for i in range(line_number+1, len(editor.lines)):
            line = editor.get_line(i)
            data = line.get_data().strip()
            if data.startswith("def "):
                break
            if data.startswith("return "):
                return True
        return False


module = {
    "class": AutoDocstring,
    "name": "autodocstring",
}



"""
Help text for Suplemon
"""

help_text = """
# Suplemon Help

*Contents*
1. General description
2. User interface
3. Default keyboard shortcuts
4. Commands

## 1. General description
Suplemon is designed to be an easy, intuitive and powerful text editor.
It emulates some features from Sublime Text and the user interface of Nano.
Multi cursor editing is a core feature. Suplemon also supports extensions
so you can customize it to work how you want.

## 2. User interface
The user interface is designed to be as intuitive and informative as possible.
There are two status bars, one at the top and one at the bottom. The top bar
shows the program version, a clock, and a list of opened files. The bottom bar
shows status messages and handles input for commands. Above the bottom status
bar there is a list of most common keyboard shortcuts.

## 3. Default keyboard shortcuts
The default keyboard shortcuts imitate those of common graphical editors.
Most shortcuts are also shown at the bottom in the legend area. Here's
the complete reference.


 * Ctrl + Q
   > Exit

 * Ctrl + W
   > Close file or tab

 * Ctrl + C
   > Copy line(s) to buffer

 * Ctrl + X
   > Cut line(s) to buffer

 * Ctrl + V
   > Insert buffer

 * Ctrl + K
   > Duplicate line

 * Ctrl + G
   > Go to line number or file (type the beginning of a filename to switch to it).
   > You can also use 'filena:42' to go to line 42 in filename.py etc.

 * Ctrl + F
   > Search for a string or regular expression (configurable)

 * Ctrl + D
   > Search for next occurance or find the word the cursor is on. Adds a new cursor at each new occurance.

 * Ctrl + T
   > Trim whitespace

 * Alt + Arrow Key
   > Add new curor in arrow direction

 * Ctrl + Left / Right
   > Jump to previous or next word or line

 * ESC
   > Revert to a single cursor / Cancel input prompt

 * Alt + Page Up
   > Move line(s) up

 * Alt + Page Down
   > Move line(s) down

 * Ctrl + S
   > Save current file

 * F1
   > Save file with new name

 * F2
   > Reload current file

 * Ctrl + O
   > Open file

 * Ctrl + W
   > Close file

 * Ctrl + Page Up
   > Switch to next file

 * Ctrl + Page Down
   > Switch to previous file

 * Ctrl + E
   > Run a command.

 * F5
   > Undo

 * F6
   > Redo

 * F7
   > Toggle visible whitespace

 * F8
   > Toggle mouse mode

 * F9
   > Toggle line numbers

 * F11
   > Toggle full screen

### Mouse shortcuts

 * Left Click
   > Set cursor at mouse position. Reverts to a single cursor.

 * Right Click
   > Add a cursor at mouse position.

 * Scroll Wheel Up / Down
   > Scroll up & down.


## 4. Commands
Commands are special operations that can be performed (e.g. remove whitespace
or convert line to uppercase). Each command can be run by pressing Ctrl + E
and then typing the command name. Commands are extensions and are stored in
the modules folder in the Suplemon installation.

 * autocomplete

    A simple autocompletion module.

    This adds autocomplete support for the tab key. It uses a word
    list scanned from all open files for completions. By default it suggests
    the shortest possible match. If there are no matches, the tab action is
    run normally.

 * autodocstring

    Simple module for adding docstring placeholders.

    This module is intended to generate docstrings for Python functions.
    It adds placeholders for descriptions, arguments and return data.
    Function arguments are crudely parsed from the function definition
    and return statements are scanned from the function body.

 * bulk_delete

    Bulk delete lines and characters.
    Asks what direction to delete in by default.

    Add 'up' to delete lines above highest cursor.
    Add 'down' to delete lines below lowest cursor.
    Add 'left' to delete characters to the left of all cursors.
    Add 'right' to delete characters to the right of all cursors.

 * comment

    Toggle line commenting based on current file syntax.

 * config

    Shortcut for openning the config files.

 * crypt

    Encrypt or decrypt the current buffer. Lets you provide a passphrase and optional salt for encryption.
    Uses AES for encryption and scrypt for key generation.

 * diff

    View a diff of the current file compared to it's on disk version.

 * eval

    Evaluate a python expression and show the result in the status bar.

    If no expression is provided the current line(s) are evaluated and
    replaced with the evaluation result.

 * keymap

    Shortcut to openning the keymap config file.

 * linter

    Linter for suplemon.

 * lower

    Transform current lines to lower case.

 * lstrip

    Trim whitespace from beginning of current lines.

 * paste

    Toggle paste mode (helpful when pasting over SSH if auto indent is enabled)

 * reload

    Reload all add-on modules.

 * replace_all

    Replace all occurrences in all files of given text with given replacement.

 * reverse

    Reverse text on current line(s).

 * rstrip

    Trim whitespace from the end of lines.

 * save

    Save the current file.

 * save_all

    Save all currently open files. Asks for confirmation.

 * sort_lines

    Sort current lines.

    Sorts alphabetically by default.
    Add 'length' to sort by length.
    Add 'reverse' to reverse the sorting.

 * strip

    Trim whitespace from start and end of lines.

 * tabstospaces

    Convert tab characters to spaces in the entire file.

 * toggle_whitespace

    Toggle visually showing whitespace.

 * upper

    Transform current lines to upper case.


"""



# -*- encoding: utf-8
"""
Base class for extension modules to inherit.
"""

import os
import json
import logging


class Storage:
    """
    Semi-automatic key/value store for suplemon modules.
    """
    def __init__(self, module):
        self.data = {}
        self.automatic = False
        self.module = module
        self.data_subdir = "modules"
        self.extension = "json"
        self.storage_dir = os.path.join(self.module.app.config.config_dir, self.data_subdir)
        if not os.path.exists(self.storage_dir):
            try:
                os.makedirs(self.storage_dir)
            except:
                self.module.app.logger.warning(
                  "Module storage folder '{0}' doesn't exist and couldn't be created.".format(
                                               self.storage_dir))

    def __getitem__(self, i):
        """Get a storage key value."""
        return self.data[i]

    def __setitem__(self, i, v):
        """Set a storage key value."""
        self.data[i] = v
        if self.automatic:
            self.store()

    def __str__(self):
        """Convert entire data dict to string."""
        return str(self.data)

    def __len__(self):
        """Return length of data dict."""
        return len(self.data)

    def keys(self):
        return self.data.keys()

    def items(self):
        return self.data.items()

    def get_path(self):
        """Get the storage file path."""
        if self.module.get_name():
            return os.path.join(self.storage_dir, self.module.get_name() + "." + self.extension)
        return False

    def get_data(self):
        """Get the storage data."""
        return self.data

    def set_data(self, data):
        """Set the storage data."""
        self.data = data

    def set_automatic(self, auto):
        """Set wether the data should be stored automatically when changed."""
        self.automatic = auto

    def store(self):
        """Store the storage data to disk."""
        try:
            data = json.dumps(self.data)
            f = open(self.get_path(), "w")
            f.write(data)
            f.close()
        except:
            self.module.logger.exception("Storing module storage failed.")

    def load(self):
        """Load the storage data from disk."""
        if os.path.exists(self.get_path()):
            try:
                f = open(self.get_path())
                data = f.read()
                f.close()
            except:
                self.module.logger.debug("Loading module storage failed.")
                return False
        else:
            return False

        try:
            self.data = json.loads(data)
            return True
        except:
            self.module.logger.exception("Parsing module storage failed.")
        return False


class Module:
    def __init__(self, app, name, options=None):
        self.app = app
        self.name = name
        self.options = options
        self.logger = None
        self.storage = Storage(self)
        self.init_logging(self.get_name())
        self.storage.load()
        self.init()

    def init(self):
        """Initialize the module.

        This function is run when the module is loaded and can be
        reimplemented for module specific initializations.
        """

    def get_name(self):
        """Get module name."""
        return self.name

    def get_options(self):
        """Get module options."""
        return self.options

    def get_status(self):
        """Called by app when to get status bar contents."""
        return ""

    def set_name(self, name):
        """Set module name."""
        self.name = name

    def set_options(self, options):
        """Set module options."""
        self.options = options

    def get_default_config(self):
        """Return module default config dict"""
        return {}

    def is_runnable(self):
        cls_method = getattr(Module, "run")
        return self.run.__module__ != cls_method.__module__

    def init_logging(self, name):
        """Initialize the module logger (self.logger).

        Should be called before the module uses logging.
        Always pass __name__ as the value to name, for consistency.

        Args:
        :param name: should be specified as __name__
        """
        if not self.logger:
            self.logger = logging.getLogger("module.{0}".format(name))

    def run(self, app, editor, args):
        """This is called each time the module is run.

        Called when command is issued via prompt or key binding.

        Args:
        :param app: the app instance
        :param editor: the current editor instance
        """
        pass

    def bind_key(self, key):
        """Shortcut for binding run method to a key.

        Args:
        :param key:
        """
        self.app.set_key_binding(key, self._proxy_run)

    def bind_event(self, event, callback):
        """Bind a callback to be called before event is run.

        If the callback returns True the event will be canceled.

        :param event: The event name
        :param callback: Function to be called
        """
        self.app.set_event_binding(event, "before", callback)

    def bind_event_before(self, event, callback):
        """Bind a callback to be called before event is run.

        If the callback returns True the event will be canceled.

        :param event: The event name
        :param callback: Function to be called
        """
        self.app.set_event_binding(event, "before", callback)

    def bind_event_after(self, event, callback):
        """Bind a callback to be called after event is run.

        :param event: The event name
        :param callback: Function to be called
        """
        self.app.set_event_binding(event, "after", callback)

    def _proxy_run(self):
        """Calls the run method with necessary arguments."""
        self.run(self.app, self.app.get_editor(), "")



# -*- encoding: utf-8

from suplemon.suplemon_module import Module


class Reload(Module):
    """Reload all add-on modules."""

    def run(self, app, editor, args):
        self.app.modules.load()


module = {
    "class": Reload,
    "name": "reload",
}



# -*- encoding: utf-8

from suplemon.suplemon_module import Module


class ReplaceAll(Module):
    """Replace all occurrences in all files of given text with given replacement."""

    def run(self, app, editor, args):
        r_from = self.app.ui.query("Replace text:")
        if not r_from:
            return False
        r_to = self.app.ui.query("Replace with:")
        if not r_to:
            return False
        for file in app.get_files():
            file.editor.replace_all(r_from, r_to)


module = {
    "class": ReplaceAll,
    "name": "replace_all",
}



# -*- encoding: utf-8

from suplemon.suplemon_module import Module


class ToggleWhitespace(Module):
    """Toggle visually showing whitespace."""

    def run(self, app, editor, args):
        # Toggle the boolean
        new_value = not self.app.config["editor"]["show_white_space"]
        self.app.config["editor"]["show_white_space"] = new_value


module = {
    "class": ToggleWhitespace,
    "name": "toggle_whitespace",
}



# -*- encoding: utf-8

from suplemon import helpers
from suplemon.suplemon_module import Module


class Comment(Module):
    """Toggle line commenting based on current file syntax."""

    def run(self, app, editor, args):
        """Comment the current line(s)."""
        try:
            # Try to get comment start and end syntax
            comment = editor.syntax.get_comment()
        except:
            return False
        line_nums = editor.get_lines_with_cursors()
        # Iterate through lines
        for lnum in line_nums:
            line = editor.lines[lnum]
            if not len(line):
                continue  # Skip empty lines
            # Look for comment syntax in stripped line (TODO:Make this smarter)
            target = str(line).strip()
            w = helpers.whitespace(line)  # Amount of whitespace at line start
            # If the line starts with comment syntax
            if target.startswith(comment[0]):
                # Reconstruct the whitespace and add the line
                new_line = (" "*w) + line[w+len(comment[0]):]
                # If comment end syntax exists
                if comment[1]:
                    # Try to remove it from the end of the line
                    if new_line.endswith(comment[1]):
                        new_line = new_line[:-1*len(comment[1])]
                # Store the modified line
                # editor.lines[lnum] = Line(new_line)
                editor.lines[lnum].set_data(new_line)
            # If the line isn't commented
            else:
                # Slice out the prepended whitespace
                new_line = line[w:]
                # Add the whitespace and starting comment
                new_line = (" "*w) + comment[0] + new_line
                if comment[1]:
                    # Add comment end syntax if needed
                    new_line += comment[1]
                # Store modified line
                # editor.lines[lnum] = Line(new_line)
                editor.lines[lnum].set_data(new_line)
        # Keep cursors under control, same as usual...
        editor.move_cursors()
        editor.store_action_state("comment")


module = {
    "class": Comment,
    "name": "comment",
}



from suplemon.linelight.color_map import color_map


class Syntax:
    def get_comment(self):
        return ("/*", "*/")

    def get_color(self, raw_line):
        color = color_map["white"]
        line = raw_line.strip()
        if line.startswith("@import"):
            color = color_map["blue"]
        elif line.startswith("$"):
            color = color_map["green"]
        elif line.startswith("/*") or line.endswith("*/"):
            color = color_map["magenta"]
        elif line.startswith("{") or line.endswith(("}", "{")):
            color = color_map["cyan"]
        elif line.endswith(";"):
            color = color_map["yellow"]
        return color



from suplemon.linelight.color_map import color_map


class Syntax:
    def get_comment(self):
        return ("# ", "")

    def get_color(self, raw_line):
        color = color_map["white"]
        line = raw_line.strip()
        keywords = ("if", "elif", "else", "finally", "try", "except",
                    "for ", "while ", "continue", "pass", "break")
        if line.startswith(("import", "from")):
            color = color_map["blue"]
        elif line.startswith("class"):
            color = color_map["green"]
        elif line.startswith("def"):
            color = color_map["cyan"]
        elif line.startswith(("return", "yield")):
            color = color_map["red"]
        elif line.startswith("self."):
            color = color_map["cyan"]
        elif line.startswith(("#", "//", "\"", "'", ":")):
            color = color_map["magenta"]
        elif line.startswith(keywords):
            color = color_map["yellow"]
        return color



# -*- encoding: utf-8

import re
import os
import threading
import subprocess

from suplemon.suplemon_module import Module


class Linter(Module):
    """Linter for suplemon."""

    def init(self):
        self.init_logging(__name__)

        # TODO: Run linting in a seperate thread to avoid
        # blocking the UI when the app is loading

        # Lint all files after app is loaded
        self.bind_event_after("app_loaded", self.on_loaded)
        # Show linting messages in status bar
        self.bind_event_after("mainloop", self.mainloop)
        # Re-lint current file when appropriate
        self.bind_event_after("save_file", self.lint_current_file)
        self.bind_event_after("save_file_as", self.lint_current_file)
        self.bind_event_after("reload_file", self.lint_current_file)
        self.bind_event_after("open_file", self.lint_current_file)

    def run(self, app, editor, args):
        """Run the linting command."""
        editor = self.app.get_file().get_editor()
        count = self.get_msg_count(editor)
        status = "{0} lines with linting errors in this file.".format(str(count))
        self.app.set_status(status)

    def on_loaded(self, event):
        # Lint all files in a thread since it might take a while
        thread = threading.Thread(target=self.lint_all_files, args=(event,))
        # Use daemon mode to forcefully shutdown the thread on application exit.
        # Python2 doesn't know the boolean daemon keyword argument to the Thread()
        # constructor so we use the .daemon property instead.
        thread.daemon = True
        thread.start()

    def mainloop(self, event):
        """Run the linting command."""
        file = self.app.get_file()
        editor = file.get_editor()
        cursor = editor.get_cursor()
        if len(editor.cursors) > 1:
            return False
        line_no = cursor.y + 1
        msg = self.get_msgs_on_line(editor, cursor.y)
        if msg:
            self.app.set_status("Line {0}: {1}".format(str(line_no), msg))

    def lint_current_file(self, event):
        self.lint_file(self.app.get_file())

    def lint_all_files(self, event):
        """Do linting check for all open files and store results."""
        for file in self.app.files:
            self.lint_file(file)
        return False

    def lint_file(self, file):
        path = file.get_path()
        if not path:  # Unsaved file
            return False

        ext = file.get_extension().lower()
        linter = False
        if ext == "py":
            linter = PyLint(self.logger)
        elif ext == "js":
            linter = JsLint(self.logger)
        elif ext == "php":
            linter = PhpLint(self.logger)

        if not linter:
            return False

        linting = linter.lint(path)
        if linting is False:
            return False

        editor = file.get_editor()
        for line_no in range(len(editor.lines)):
            line = editor.lines[line_no]
            if line_no+1 in linting.keys():
                line.linting = linting[line_no+1]
                line.set_number_color(1)
            else:
                line.linting = False
                line.reset_number_color()

    def get_msgs_on_line(self, editor, line_no):
        line = editor.lines[line_no]
        if not hasattr(line, "linting") or not line.linting:
            return False
        return line.linting[0][1]

    def get_msg_count(self, editor):
        count = 0
        for line in editor.lines:
            if hasattr(line, "linting"):
                if line.linting:
                    count += 1
        return count


class BaseLint:
    def __init__(self, logger):
        self.logger = logger

    def lint(self, path):
        pass

    def get_file_linting(self, path):
        """Do linting check for given file path."""
        return {}

    def get_output(self, cmd):
        try:
            fnull = open(os.devnull, "w")
            process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=fnull)
            fnull.close()
        except (OSError, EnvironmentError):  # can't use FileNotFoundError in Python 2
            self.logger.debug("Subprocess failed.")
            return False
        out, err = process.communicate()
        return out


class PyLint(BaseLint):
    def __init__(self, logger):
        BaseLint.__init__(self, logger)
        # Error codes to ignore e.g. 'E501' (line too long)
        self.ignore = []
        # Max length of line
        self.max_line_length = 120  # Default is 79

    def lint(self, path):
        if not self.has_flake8_support():
            self.logger.warning("Flake8 not available. Can't show linting.")
            return False
        self.logger.debug(self.has_flake8_support())
        return self.get_file_linting(path)

    def has_flake8_support(self):
        output = self.get_output(["flake8", "--version"])
        return output

    def get_file_linting(self, path):
        """Do linting check for given file path."""
        output = self.get_output(["flake8", "--max-line-length", str(self.max_line_length), path])
        if output is False:
            self.logger.warning("Failed to get linting for file '{0}'.".format(path))
            return False
        output = output.decode("utf-8")
        # Remove file paths from output
        output = output.replace(path+":", "")
        lines = output.split("\n")
        linting = {}
        for line in lines:
            if not line:
                continue
            try:
                parts = line.split(":")
                line_no = int(parts[0])
                char_no = int(parts[1])
                data = ":".join(parts[2:]).strip()
                err_code = data.split(" ")[0]
                if err_code in self.ignore:
                    continue
                if line_no not in linting.keys():
                    linting[line_no] = []
                linting[line_no].append((char_no, data, err_code))
            except:
                self.logger.debug("Failed to parse line:{0}".format(line))
        return linting


class JsLint(BaseLint):
    def __init__(self, logger):
        BaseLint.__init__(self, logger)

    def lint(self, path):
        return self.get_file_linting(path)

    def get_file_linting(self, path):
        """Do linting check for given file path."""
        output = self.get_output(["jshint", path])
        if output is False:
            self.logger.warning("Failed to get linting for file '{0}'.".format(path))
            return False
        output = output.decode("utf-8")
        # Remove file paths from output
        output = output.replace(path+": ", "")
        lines = output.split("\n")
        linting = {}
        for line in lines:
            if not line:
                continue
            try:
                parts = line.split(", ")
                if len(parts) < 3:
                    continue
                line_no = int(re.sub("\D", "", parts[0]))
                char_no = int(re.sub("\D", "", parts[1]))
                data = parts[2]
                err_code = None
                if line_no not in linting.keys():
                    linting[line_no] = []
                linting[line_no].append((char_no, data, err_code))
            except:
                self.logger.debug("Failed to parse line:{0}".format(line))
        return linting


class PhpLint(BaseLint):
    def __init__(self, logger):
        BaseLint.__init__(self, logger)

    def lint(self, path):
        return self.get_file_linting(path)

    def get_output(self, cmd, errors=False):
        # Need to override this since php reports the errors to stderr
        try:
            fnull = open(os.devnull, "w")
            process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            fnull.close()
        except (OSError, EnvironmentError):  # can't use FileNotFoundError in Python 2
            self.logger.debug("Subprocess failed.")
            return False
        out, err = process.communicate()
        return err

    def get_file_linting(self, path):
        """Do linting check for given file path."""
        output = self.get_output(["php", "-l", path])
        if output is False:
            self.logger.warning("Failed to get linting for file '{0}'.".format(path))
            return False
        output = output.decode("utf-8")
        lines = output.split("\n")
        linting = {}
        for line in lines:
            line = line.strip()
            if not line:
                continue
            try:
                parts = line.split(" in {0} on line ".format(path))
                self.logger.debug(parts)
                if len(parts) != 2:
                    continue
                line_no = int(parts[1])
                char_no = 0
                data = parts[0]
                err_code = None
                if line_no not in linting.keys():
                    linting[line_no] = []
                linting[line_no].append((char_no, data, err_code))
            except:
                self.logger.debug("Failed to parse line:{0}".format(line))
        return linting


module = {
    "class": Linter,
    "name": "linter",
}



# -*- encoding: utf-8
"""
File object for storing an opened file and editor.
"""

import os
import time
import logging

from .helpers import parse_path


class File:
    def __init__(self, app=None):
        self.app = app
        self.logger = logging.getLogger(__name__)
        self.name = ""
        self.fpath = ""
        self.data = None
        self.read_only = False  # Currently unused
        self.last_save = None  # Time of last save
        self.opened = time.time()  # Time of last open
        self.editor = None
        self.writable = True
        self.is_help = False

    def _path(self):
        """Get the full path of the file."""
        return os.path.join(self.fpath, self.name)

    def path(self):
        """Get the full path of the file."""
        # TODO: deprecate in favour of get_path()
        return self._path()

    def get_name(self):
        """Get the file name."""
        return self.name

    def get_path(self):
        """Get the full path of the file."""
        return self._path()

    def get_extension(self):
        parts = self.name.split(".")
        if len(parts) < 2:
            return ""
        return parts[-1]

    def get_editor(self):
        """Get the associated editor."""
        return self.editor

    def set_name(self, name):
        """Set the file name."""
        # TODO: sanitize
        self.name = name
        self.update_editor_extension()

    def set_path(self, path):
        """Set the file path. Relative paths are sanitized."""
        self.fpath, self.name = parse_path(path)
        self.update_editor_extension()

    def set_data(self, data):
        """Set the file data and apply to editor if it exists."""
        self.data = data
        if self.editor:
            self.editor.set_data(data)

    def set_editor(self, editor):
        """The editor instance set its file extension."""
        self.editor = editor
        self.update_editor_extension()

    def on_load(self):
        """Does checks after file is loaded."""
        self.writable = os.access(self._path(), os.W_OK)
        if not self.writable:
            self.logger.info("File not writable.")

    def update_editor_extension(self):
        """Set the editor file extension from the current file name."""
        if not self.editor:
            return False
        ext = self.get_extension()
        if len(ext) >= 1:
            self.editor.set_file_extension(ext)

    def save(self):
        """Write the editor data to file."""
        data = self.editor.get_data()
        try:
            f = open(self._path(), "w")
            f.write(data)
            f.close()
        except:
            return False
        self.data = data
        self.last_save = time.time()
        self.writable = os.access(self._path(), os.W_OK)
        return True

    def load(self, read=True):
        """Try to read the actual file and load the data into the editor instance."""
        if not read:
            return True
        path = self._path()
        if not os.path.isfile(path):
            self.logger.debug("Given path isn't a file.")
            return False
        data = self._read(path)
        if data is False:
            return False
        self.data = data
        self.editor.set_data(data)
        self.on_load()
        return True

    def _read(self, path):
        data = self._read_text(path)
        if data is False:
            self.logger.warning("Normal file read failed.")
            data = self._read_binary(path)
        if data is False:
            self.logger.warning("Fallback file read failed.")
            return False
        return data

    def _read_text(self, file):
        # Read text file
        try:
            f = open(self._path())
            data = f.read()
            f.close()
            return data
        except:
            self.logger.exception("Failed reading file \"{file}\"".format(file=file))
            return False

    def _read_binary(self, file):
        # Read binary file and try to autodetect encoding
        try:
            f = open(self._path(), "rb")
            data = f.read()
            f.close()
            import chardet
            detection = chardet.detect(data)
            charenc = detection["encoding"]
            if charenc is None:
                self.logger.warning("Failed to detect file encoding.")
                return False
            self.logger.info("Trying to decode with encoding '{0}'".format(charenc))
            return data.decode(charenc)
        except:
            self.logger.warning("Failed reading binary file!", exc_info=True)
        return False

    def reload(self):
        """Reload file data."""
        return self.load()

    def is_changed(self):
        """Check if the editor data is different from the file."""
        return self.editor.get_data() != self.data

    def is_changed_on_disk(self):
        path = self._path()
        if os.path.isfile(path):
            data = self._read(path)
            if data != self.data:
                return True
        return False

    def is_writable(self):
        """Check if the file is writable."""
        return self.writable



# -*- encoding: utf-8

from suplemon.suplemon_module import Module


class Paste(Module):
    """Toggle paste mode (helpful when pasting over SSH if auto indent is enabled)"""

    def init(self):
        # Flag for paste mode
        self.active = False
        # Initial state of auto indent
        self.auto_indent_active = self.app.config["editor"]["auto_indent_newline"]
        # Listen for config changes
        self.bind_event_after("config_loaded", self.config_loaded)

    def run(self, app, editor, args):
        # Simply toggle pastemode when the command is run
        self.active = not self.active
        self.set_paste_mode(self.active)
        self.show_confirmation()
        return True

    def config_loaded(self, e):
        # Refresh the auto indent state when config is reloaded
        self.auto_indent_active = self.app.config["editor"]["auto_indent_newline"]

    def get_status(self):
        # Return the paste mode status for the statusbar
        return "[PASTEMODE]" if self.active else ""

    def show_confirmation(self):
        # Show a status message when pastemode is toggled
        state = "activated" if self.active else "deactivated"
        self.app.set_status("Paste mode " + state)

    def set_paste_mode(self, active):
        # Enable or disable auto indent
        if active:
            self.app.config["editor"]["auto_indent_newline"] = False
        else:
            self.app.config["editor"]["auto_indent_newline"] = self.auto_indent_active


module = {
    "class": Paste,
    "name": "paste",
    "status": "bottom",
}



# -*- encoding: utf-8

import os
import time
import subprocess

from suplemon import helpers
from suplemon.suplemon_module import Module


class Battery(Module):
    """Shows remaining battery capacity in the top status bar if available."""

    def init(self):
        self.last_value = -1
        self.checked = time.time()
        self.interval = 60  # Seconds to wait until polling again

    def value(self):
        """Get the battery charge percent and cache it."""
        if self.last_value == -1:
            state = self.battery_status()
        elif time.time()-self.checked > self.interval:
            state = self.battery_status()
        else:
            return self.last_value
        self.last_value = state
        return state

    def value_str(self):
        """Return formatted value string to show in the UI."""
        val = self.value()
        if val:
            if self.app.config["app"]["use_unicode_symbols"]:
                return "\u26A1{0}%".format(str(val))
            else:
                return "BAT {0}%".format(str(val))
        return ""

    def get_status(self):
        """Called by app when showing status bar contents."""
        return self.value_str()

    def battery_status(self):
        """Attempts to get the battery charge percent."""
        value = None
        methods = [
            self.battery_status_read,
            self.battery_status_acpi,
            self.battery_status_upower
        ]
        for m in methods:
            value = m()
            if value is not None:
                break
        return value

    def battery_status_read(self):
        """Get the battery status via proc/acpi."""
        try:
            path_info = self.readf("/proc/acpi/battery/BAT0/info")
            path_state = self.readf("/proc/acpi/battery/BAT0/state")
        except:
            return None
        try:
            max_cap = float(helpers.get_string_between("last full capacity:", "mWh", path_info))
            cur_cap = float(helpers.get_string_between("remaining capacity:", "mWh", path_state))
            return int(cur_cap / max_cap * 100)
        except:
            return None

    def battery_status_acpi(self):
        """Get the battery status via acpi."""
        try:
            fnull = open(os.devnull, "w")
            raw_str = subprocess.check_output(["acpi"], stderr=fnull)
            fnull.close()
        except:
            return None
        raw_str = raw_str.decode("utf-8")
        part = helpers.get_string_between(",", "%", raw_str)
        if part:
            try:
                return int(part)
            except:
                return None
        return None

    def battery_status_upower(self):
        """Get the battery status via upower."""
        path = "/org/freedesktop/UPower/devices/battery_BAT0"
        try:
            raw_str = subprocess.check_output(["upower", "-i", path])
        except:
            return None
        raw_str = raw_str.decode("utf-8")
        raw_str = raw_str.splitlines()[0]
        part = helpers.get_string_between("percentage:", "%", raw_str)
        if part:
            try:
                return int(part)
            except:
                return None
        return None

    def readf(self, path):
        """Read and return file contents at path."""
        f = open(path)
        data = f.read()
        f.close()
        return data


module = {
    "class": Battery,
    "name": "battery",
    "status": "top",
}



# -*- encoding: utf-8
"""
Theme loader
"""
import os
import logging

try:
    import xml.etree.cElementTree as ET
except:
    import xml.etree.ElementTree as ET

from . import hex2xterm

# Map scope name to its color pair index
scope_to_pair = {
    "global": 21,
    "comment": 22,
    "string": 23,
    "constant.numeric": 24,
    "constant.language": 25,
    "constant.character": 26,
    "constant.other": 27,
    "variable": 28,
    "keyword": 29,
    "storage": 30,
    "storage.type": 31,
    "entity.name.class": 32,
    "entity.other.inherited-class": 33,
    "entity.name.function": 34,
    "variable.parameter": 35,
    "entity.name.tag": 36,
    "entity.other.attribute-name": 37,
    "support.function": 38,
    "support.constant": 39,
    "support.type": 40,
    "support.class": 41,
    "support.other.variable": 42,
    "invalid": 43,
    "invalid.deprecated": 44,
    "meta.structure.dictionary.json string.quoted.double.json": 45,
    "meta.diff": 46,
    "meta.diff.header": 47,
    "markup.deleted": 48,
    "markup.inserted": 49,
    "markup.changed": 50,
    "constant.numeric.line-number.find-in-files - match": 51,
    "entity.name.filename.find-in-files": 52,
}


class Theme:
    def __init__(self, name, uuid):
        self.name = name
        self.uuid = uuid
        self.scopes = {}


class ThemeLoader:
    def __init__(self, app=None):
        self.app = app
        self.logger = logging.getLogger(__name__)

        self.curr_path = os.path.dirname(os.path.realpath(__file__))
        # The themes subdirectory
        self.theme_path = os.path.join(self.curr_path, "themes" + os.sep)
        # Module instances
        self.themes = {}
        self.current_theme = None

    def load(self, name):
        fullpath = "".join([self.theme_path, name, ".tmTheme"])
        if not os.path.exists(fullpath):
            self.logger.warning("fullpath '{0}' doesn't exist!".format(fullpath))
            return None

        self.logger.debug("Loading theme '{0}'.".format(name))
        try:
            tree = ET.parse(fullpath)
        except:
            self.logger.error("Couldn't parse theme '{}'. Falling back to line based highlighting.".format(name))
            return None
        root = tree.getroot()

        for child in root:
            config = self.parse(child)
            if config is None:
                return None

        name = config.get("name")
        uuid = config.get("uuid")

        theme = Theme(name, uuid)

        try:
            settings = config["settings"]
        except:
            return None
        self.set_theme(theme, settings)

        return theme

    def use(self, name):
        theme = None
        try:
            theme = self.themes[name]
        except:
            theme = self.load(name)
            self.themes[name] = theme
        if theme is None:
            return
        self.current_theme = theme

    def get_scope(self, name):
        if self.current_theme:
            return self.current_theme.scopes.get(name)
        return False

    def set_theme(self, theme, settings):
        for entry in settings:
            if not isinstance(entry, dict):
                continue
            scope_str = entry.get("scope") or "global"
            scopes = scope_str.split(",")
            settings = entry.get("settings")
            if settings is not None:
                for scope in scopes:
                    theme.scopes[scope.strip()] = settings

    def parse(self, node):
        if node.tag == "dict":
            return self.parse_dict(node)
        elif node.tag == "array":
            return self.parse_array(node)
        elif node.tag == "string":
            return self.parse_text(node)

        return None

    def parse_dict(self, node):
        d = {}
        key = None
        value = None

        for child in node:
            if key is None:
                if child.tag != "key":  # key expected
                    return None
                key = child.text
            else:
                value = self.parse(child)
                if value is not None:
                    d[key] = value

                key = None
                value = None

        return d

    def parse_array(self, node):
        l = []
        for child in node:
            value = self.parse(child)
            if value is not None:
                l.append(value)

        return l

    def parse_text(self, node):
        # If the node text is a hex color convert it to an xterm equivalent
        if node.text:
            color = self.convert_color(node.text)
            if color is not False:
                return color

        return node.text

    def convert_color(self, text):
        if text[0] != "#":
            return False
        # Validate length e.g. #F90, #FF9900, #FF990055
        if len(text) not in [4, 7, 9]:
            return False
        # Strip alpha channel
        hex_color = text
        if len(hex_color) == 9:
            hex_color = hex_color[:7]
        # If the color only has one character per channel convert them to two
        if len(hex_color) == 4:
            hex_color = "#{}{}{}".format(hex_color[1:2]*2, hex_color[2:3]*2, hex_color[3:4]*2)
        # Convert the color
        try:
            xterm_color = int(hex2xterm.hex_to_xterm(hex_color))
            return xterm_color
        except:
            self.logger.exception("Failed to convert color: {}".format(text))

        return False



# -*- encoding: utf-8
"""
Cursor object for storing cursor data.
"""


class Cursor:
    def __init__(self, x=0, y=0):
        """Initialize Cursor.

        :param x: Cursor x coordinate or x,y tuple. Defaults to 0.
        :param y: Cursor y coordinate. Defaults to 0.
        """
        # Handle coords as a tuple
        if isinstance(x, tuple) or isinstance(x, list):
            x, y = x
            self.x = x
            self.y = y
        # Handle coords from existing Cursor
        elif isinstance(x, Cursor):  # Handle arguments as a cursor
            self.x = x.x
            self.y = x.y
        # Handle coords as plain ints
        else:
            self.x = x
            self.y = y
        # Store the desired x position and
        # use it if the line is long enough
        self.persistent_x = self.x

    def get_x(self):
        """Return the x coordinate of the cursor.

        :return: Cursor x coordinate.
        :rtype: int
        """
        return self.x

    def get_y(self):
        """Return the y coordinate of the cursor.

        :return: Cursor y coordinate.
        :rtype: int
        """
        return self.y

    def set_x(self, x):
        """Set the x coordinate of the cursor."""
        self.x = x
        self.persistent_x = x

    def set_y(self, y):
        """Set the y coordinate of the cursor."""
        self.y = y

    def move_left(self, delta=1):
        """Move the cursor left by delta steps.

        :param int delta: How much to move. Defaults to 1.
        """
        self.x -= delta
        if self.x < 0:
            self.x = 0
        self.persistent_x = self.x
        return

    def move_right(self, delta=1):
        """Move the cursor right by delta steps.

        :param int delta: How much to move. Defaults to 1.
        """
        self.x += delta
        # Check in case of negative values
        if self.x < 0:
            self.x = 0
        self.persistent_x = self.x
        return

    def move_up(self, delta=1):
        """Move the cursor up by delta steps.

        :param int delta: How much to move. Defaults to 1.
        """
        self.y -= delta
        if self.y < 0:
            self.y = 0
        return

    def move_down(self, delta=1):
        """Move the cursor down by delta steps.

        :param int delta: How much to move. Defaults to 1.
        """
        self.y += delta
        return

    def __getitem__(self, i):
        # TODO: Deprecate in favor of proper access methods.
        """Get coordinates with list indices.

        :param i: 0 or 1 for x or y
        :return: x or y coordinate of cursor.
        :rtype: int
        """
        if i == 0:
            return self.x
        elif i == 1:
            return self.y

    def __eq__(self, item):
        """Check cursor for equality."""
        if isinstance(item, Cursor):
            if item.x == self.x and item.y == self.y:
                return True
        return False

    def __ne__(self, item):
        """Check cursor for unequality."""
        if isinstance(item, Cursor):
            if item.x != self.x or item.y != self.x:
                return False

    def __str__(self):
        return "Cursor({x},{y})".format(x=self.x, y=self.y)

    def __repr__(self):
        return self.__str__()

    def tuple(self):
        """Return the cursor as a tuple.

        :return: Tuple with x and y coordinates of cursor.
        :rtype: tuple
        """
        return (self.x, self.y)



from suplemon.linelight.color_map import color_map


class Syntax:
    def get_comment(self):
        return ("<!--", "-->")

    def get_color(self, raw_line):
        color = color_map["white"]
        line = raw_line.strip()
        if line.startswith(("#", "//", "/*", "*/", "<!--")):
            color = color_map["magenta"]
        elif line.endswith(("*/", "-->")):
            color = color_map["magenta"]
        elif line.startswith("<"):
            color = color_map["cyan"]
        elif line.endswith(">"):
            color = color_map["cyan"]
        return color



# -*- encoding: utf-8
"""
Editor class for extending viewer with text editing features.
"""


from . import helpers

from .line import Line
from .cursor import Cursor
from .viewer import Viewer


class State:
    """Store editor state for undo/redo."""
    def __init__(self, editor=None):
        self.cursors = [Cursor()]
        self.lines = [Line()]
        self.y_scroll = 0
        self.x_scroll = 0
        self.last_find = ""
        if editor is not None:
            self.store(editor)

    def store(self, editor):
        """Store the state of editor instance."""
        self.cursors = [cursor.tuple() for cursor in editor.cursors]
        self.lines = [line.data for line in editor.lines]
        self.y_scroll = editor.y_scroll
        self.x_scroll = editor.x_scroll
        self.last_find = editor.last_find

    def restore(self, editor):
        """Restore stored state into the editor instance."""
        editor.cursors = [Cursor(cursor) for cursor in self.cursors]
        editor.lines = [Line(line) for line in self.lines]
        editor.y_scroll = self.y_scroll
        editor.x_scroll = self.x_scroll
        editor.last_find = self.last_find


class Editor(Viewer):
    """Extends Viewer with editing capabilities."""
    def __init__(self, app, window):
        """Initialize the editor.

        Args:
            app: The Suplemon main instance.
            window: A window object to use for the ui.
        """
        Viewer.__init__(self, app, window)

        # History of editor states for undo/redo
        self.history = [State()]
        # Current state index of the editor
        self.current_state = 0
        # Last editor action that was used (for undo/redo)
        self.last_action = None

    def init(self):
        Viewer.init(self)
        operations = {
            "backspace": self.backspace,                  # Backspace
            "delete": self.delete,                        # Delete
            "insert": self.insert,                        # Insert
            "enter": self.enter,                          # Enter
            "tab": self.tab,                              # Tab
            "untab": self.untab,                          # Shift + Tab
            "escape": self.escape,                        # Escape
            "single_selection": self.single_selection,    # Escape
            "clear_last_find": self.clear_last_find,      # Escape
            "new_cursor_up": self.new_cursor_up,          # Alt + Up
            "new_cursor_down": self.new_cursor_down,      # Alt + Down
            "new_cursor_left": self.new_cursor_left,      # Alt + Left
            "new_cursor_right": self.new_cursor_right,    # Alt + Right
            "page_up": self.page_up,                      # Page Up
            "page_down": self.page_down,                  # Page Down
            "push_up": self.push_up,                      # Alt + Page Up
            "push_down": self.push_down,                  # Alt + Page Down
            "undo": self.undo,                            # F5
            "redo": self.redo,                            # F6
            "toggle_line_nums": self.toggle_line_nums,    # F9
            "toggle_line_ends": self.toggle_line_ends,    # F10
            "toggle_highlight": self.toggle_highlight,    # F11
            "copy": self.copy,                            # Ctrl + C
            "cut": self.cut,                              # Ctrl + X
            "duplicate_line": self.duplicate_line,        # Ctrl + W
        }
        for key in operations.keys():
            self.operations[key] = operations[key]

    def set_buffer(self, buffer):
        """Sets local or global buffer depending on config."""
        if self.app.config["editor"]["use_global_buffer"]:
            self.app.global_buffer = buffer
        else:
            self.buffer = buffer

    def set_data(self, data):
        """Set the editor text contents."""
        Viewer.set_data(self, data)
        buffer = self.get_buffer()  # TODO: check this
        if len(buffer) > 1:
            self.store_state()
        else:
            state = State()
            state.store(self)
            self.history[0] = state

    def store_action_state(self, action, state=None):
        """Store the editor state if a new action is taken."""
        if self.last_action != action:
            self.last_action = action
            self.store_state(state)
        else:
            # FIXME: This if is here just for safety.
            # FIXME: current_state might be wrong ;.<
            if self.current_state < len(self.history)-1:
                self.history[self.current_state].store(self)

    def store_state(self, state=None, action=None):
        """Store the current editor state for undo/redo."""
        if state is None:
            state = State()
            state.store(self)
        if len(self.history) > 1:
            if self.current_state < len(self.history)-1:
                self.history = self.history[:self.current_state]

        self.history.append(state)
        self.current_state = len(self.history)-1

        if len(self.history) > self.config["max_history"]:
            self.history.pop(0)

    def restore_state(self, index=None):
        """Restore an editor state."""
        if len(self.history) <= 1:
            return False
        if index is None:
            index = self.current_state-1

        if index < 0 or index >= len(self.history):
            return False

        # if self.current_state < len(self.history):
        #     self.current_state = self.current_state-1

        state = self.history[index]
        state.restore(self)
        self.current_state = index

    def handle_input(self, event):
        done = Viewer.handle_input(self, event)
        if not done:
            if event.is_typeable:
                if isinstance(event.key_code, str):
                    self.type(event.key_code)
                elif event.key_name:
                    self.type(event.key_name)
                return True
        return False

    def undo(self):
        """Undo the last command or change."""
        self.last_action = "undo"
        self.restore_state()

    def redo(self):
        """Redo the last command or change."""
        self.last_action = "redo"
        if self.current_state == len(self.history)-1:
            return False
        index = self.current_state+1
        self.restore_state(index)

    #
    # Cursor operations
    #

    def new_cursor_up(self):
        """Add a new cursor one line up."""
        x = self.get_cursor().x
        cursor = self.get_first_cursor()
        if cursor.y == 0:
            return
        new = Cursor(x, cursor.y-1)
        self.cursors.append(new)
        self.move_cursors()
        self.scroll_up()

    def new_cursor_down(self):
        """Add a new cursor one line down."""
        x = self.get_cursor().x
        cursor = self.get_last_cursor()
        if cursor.y == len(self.lines)-1:
            return
        new = Cursor(x, cursor.y+1)
        self.cursors.append(new)
        self.move_cursors()
        self.scroll_down()

    def new_cursor_left(self):
        """Add a new cursor one character left."""
        new = []
        for cursor in self.cursors:
            if cursor.x == 0:
                continue
            new.append(Cursor(cursor.x-1, cursor.y))
        for c in new:
            self.cursors.append(c)
        self.move_cursors()
        self.scroll_up()

    def new_cursor_right(self):
        """Add a new cursor one character right."""
        new = []
        for cursor in self.cursors:
            if cursor.x+1 > len(self.lines[cursor.y]):
                continue
            new.append(Cursor(cursor.x+1, cursor.y))
        for c in new:
            self.cursors.append(c)
        self.move_cursors()
        self.scroll_down()

    def escape(self):
        """Handle escape key.

        Wrapper for clear_last_find and single_selection."""
        self.clear_last_find()
        self.single_selection()

    def clear_last_find(self):
        """Removes last_find so a new auto-find can be initiated."""
        self.last_find = ""

    def single_selection(self):
        """Removes all cursors except primary cursor."""
        self.cursors = [self.cursors[0]]
        self.move_cursors()

    #
    # Text editing operations
    #

    def replace_all(self, what, replacement):
        """Replaces what with replacement on each line."""
        for line in self.lines:
            data = line.get_data()
            new = data.replace(what, replacement)
            line.set_data(new)
        self.move_cursors()

    def delete(self):
        """Delete the next character."""
        for cursor in self.cursors:
            if len(self.lines)-1 < cursor.y:
                # If we've run out of lines
                break
            line = self.lines[cursor.y]
            # if we have more than 1 line
            # and we're at the end of the current line
            # and we're not on the last line
            if len(self.lines) > 1 and cursor.x == len(line) and cursor.y != len(self.lines) - 1:
                data = self.lines[cursor.y].get_data()
                self.lines.pop(cursor.y)
                self.lines[cursor.y].set_data(data+self.lines[cursor.y])
                # Reposition cursors from line below into correct positions on current line
                line_cursors = self.get_cursors_on_line(cursor.y+1)
                for c in line_cursors:
                    c.move_right(len(data))
                    c.move_up()
                self.move_y_cursors(cursor.y, -1)
            else:
                start = line[:cursor.x]
                end = line[cursor.x+1:]
                self.lines[cursor.y].set_data(start+end)
                self.move_x_cursors(cursor.y, cursor.x, -1)
        self.move_cursors()
        # Add a restore point if previous action != delete
        self.store_action_state("delete")

    def backspace(self):
        """Delete the previous character."""
        curs = reversed(sorted(self.cursors, key=lambda c: (c[1], c[0])))
        # Iterate through all cursors from bottom to top
        for cursor in curs:
            line_no = cursor.y
            # If we're at the beginning of file don't do anything
            if cursor.x == 0 and cursor.y == 0:
                continue
            # If were operating at the beginning of a line
            if cursor.x == 0 and cursor.y != 0:
                curr_line = self.lines.pop(line_no)
                prev_line = self.lines[line_no-1]
                length = len(prev_line)  # Get the length of previous line

                # Add the current line to the previous one
                new_data = self.lines[cursor.y-1] + curr_line
                self.lines[cursor.y-1].set_data(new_data)

                # Get all cursors on current line
                line_cursors = self.get_cursors_on_line(line_no)

                for line_cursor in line_cursors:  # Move the cursors
                    line_cursor.move_up()
                    # Add the length of previous line to each x coordinate
                    # so that their relative positions
                    line_cursor.move_right(length)
                # Move all cursors below up one line
                # (since a line was removed above them)
                self.move_y_cursors(cursor.y, -1)
            # Handle all other cases
            else:
                curr_line = self.lines[line_no]
                # Remove one character by default
                del_n_chars = 1
                # Check if we should unindent
                if self.config["backspace_unindent"]:
                    # Check if we can unindent, and that it's actually whitespace
                    # We don't do this for hard tabs since they're just a single character
                    if not self.config["hard_tabs"]:
                        indent = self.config["tab_width"]
                        if cursor.x >= indent:
                            if curr_line[cursor.x-indent:cursor.x] == indent*" ":
                                # Remove an indents worth of whitespace
                                del_n_chars = indent
                # Slice characters out of the line
                start = curr_line[:cursor.x-del_n_chars]
                end = curr_line[cursor.x:]
                # Store the new line
                self.lines[line_no].set_data(start+end)
                # Move the operating curser back the deleted amount
                cursor.move_left(del_n_chars)
                # Do the same to the rest
                self.move_x_cursors(line_no, cursor.x, -1*del_n_chars)
        # Ensure we keep the view scrolled
        self.move_cursors()
        self.scroll_up()
        # Add a restore point if previous action != backspace
        self.store_action_state("backspace")

    def enter(self):
        """Insert a new line at each cursor."""
        # We sort the cursors, and loop through them from last to first
        # That way we avoid messing with
        # the relative positions of the higher cursors
        curs = sorted(self.cursors, key=lambda c: (c[1], c[0]))
        curs = reversed(curs)
        for cursor in curs:
            # The current line this cursor is on
            line = self.lines[cursor.y]

            # Start of the line
            start = line[:cursor.x]

            # End of the line
            end = line[cursor.x:]

            # Leave the beginning of the line
            self.lines[cursor.y].set_data(start)
            wspace = ""
            if self.config["auto_indent_newline"]:
                wspace = helpers.whitespace(self.lines[cursor.y])*" "
            self.lines.insert(cursor.y+1, Line(wspace+end))
            self.move_y_cursors(cursor.y, 1)
            cursor.set_x(len(wspace))
            cursor.move_down()
        self.move_cursors()
        self.scroll_down()
        # Add a restore point if previous action != enter
        self.store_action_state("enter")

    def insert(self):
        """Insert buffer data at cursor(s)."""
        cur = self.get_cursor()
        buffer = list(self.get_buffer())

        # If we have more than one cursor
        # Or one cursor and one line
        if len(self.cursors) > 1 or len(buffer) == 1:
            # If the cursor count is more than the buffer length extend
            # the buffer until it's at least as long as the cursor count
            while len(buffer) < len(self.cursors):
                buffer.extend(buffer)
            curs = sorted(self.cursors, key=lambda c: (c[1], c[0]))
            for cursor in curs:
                line = self.lines[cursor.y]
                buf = buffer[0]
                line = line[:cursor.x] + buf + line[cursor.x:]
                self.lines[cursor.y].set_data(line)
                buffer.pop(0)
                self.move_x_cursors(cursor.y, cursor.x-1, len(buf))
        # If we have one cursor and multiple lines
        else:
            for buf in buffer:
                y = cur[1]
                if y < 0:
                    y = 0
                self.lines.insert(y, Line(buf))
                self.move_y_cursors(cur[1]-1, 1)
        self.move_cursors()
        self.scroll_down()
        # Add a restore point if previous action != insert
        self.store_action_state("insert")

    def insert_lines_at(self, lines, at):
        rev_lines = reversed(lines)
        for line in rev_lines:
            self.lines.insert(at, Line(line))
        self.move_y_cursors(at, len(lines))

    def push_up(self):
        """Move current lines up by one line."""
        used_y = []
        curs = sorted(self.cursors, key=lambda c: (c[1], c[0]))
        for cursor in curs:
            if cursor.y in used_y:
                continue
            used_y.append(cursor.y)
            if cursor.y == 0:
                break
            old = self.lines[cursor.y-1]
            self.lines[cursor.y-1] = self.lines[cursor.y]
            self.lines[cursor.y] = old
        self.move_cursors((0, -1))
        self.scroll_up()
        # Add a restore point if previous action != push_up
        self.store_action_state("push_up")

    def push_down(self):
        """Move current lines down by one line."""
        used_y = []
        curs = reversed(sorted(self.cursors, key=lambda c: (c[1], c[0])))
        for cursor in curs:
            if cursor.y in used_y:
                continue
            if cursor.y >= len(self.lines)-1:
                break
            used_y.append(cursor.y)
            old = self.lines[cursor.y+1]
            self.lines[cursor.y+1] = self.lines[cursor.y]
            self.lines[cursor.y] = old

        self.move_cursors((0, 1))
        self.scroll_down()
        # Add a restore point if previous action != push_down
        self.store_action_state("push_down")

    def tab(self):
        """Indent lines."""
        # Add a restore point if previous action != tab
        self.store_action_state("tab")
        if not self.config["hard_tabs"]:
            self.type(" "*self.config["tab_width"])
        else:
            self.type("\t")

    def untab(self):
        """Unindent lines."""
        linenums = []
        # String to compare tabs to
        tab = " "*self.config["tab_width"]
        if self.config["hard_tabs"]:
            tab = "\t"
        width = len(tab)
        for cursor in self.cursors:
            line = self.lines[cursor.y]
            if cursor.y in linenums:
                cursor.x = helpers.whitespace(line)
                continue
            elif line[:width] == tab:
                line = Line(line[width:])
                self.lines[cursor.y] = line
                cursor.x = helpers.whitespace(line)
                linenums.append(cursor.y)
        # Add a restore point if previous action != untab
        self.store_action_state("untab")

    def copy(self):
        """Copy lines to buffer."""
        # Store cut lines in buffer
        copy_buffer = []
        # Get all lines with cursors on them
        line_nums = self.get_lines_with_cursors()

        for i in range(len(line_nums)):
            # Get the line
            line = self.lines[line_nums[i]]
            # Put it in our temporary buffer
            copy_buffer.append(line.get_data())
        self.set_buffer(copy_buffer)
        self.store_action_state("copy")

    def cut(self):
        """Cut lines to buffer."""
        # Store cut lines in buffer
        cut_buffer = []
        # Get all lines with cursors on them
        line_nums = self.get_lines_with_cursors()
        # Sort from last to first (invert order)
        line_nums = line_nums[::-1]
        for i in range(len(line_nums)):  # Iterate from last to first
            # Make sure we don't completely remove the last line
            if len(self.lines) == 1:
                cut_buffer.append(self.lines[0])
                self.lines[0] = Line()
                break
            # Get the current line
            line_no = line_nums[i]
            # Get and remove the line
            line = self.lines.pop(line_no)
            # Put it in our temporary buffer
            cut_buffer.append(line)
            # Move all cursors below the current line up
            self.move_y_cursors(line_no, -1)
        self.move_cursors()  # Make sure cursors are in valid places
        # Reverse the buffer to get correct order and store it
        self.set_buffer(cut_buffer[::-1])
        self.store_action_state("cut")

    def type(self, data):
        """Insert data at each cursor position."""
        for cursor in self.cursors:
            self.type_at_cursor(cursor, data)
        self.move_cursors()
        # Add a restore point if previous action != type
        self.store_action_state("type")

    def type_at_cursor(self, cursor, data):
        """Insert data at specified cursor."""
        line = self.lines[cursor.y]
        start = line[:cursor.x]
        end = line[cursor.x:]
        self.lines[cursor.y].set_data(start + data + end)
        self.move_x_cursors(cursor.y, cursor.x, len(data))
        cursor.move_right(len(data))

    def go_to_pos(self, line_no, col=0):
        """Move primary cursor to line_no, col=0."""
        if line_no < 0:
            line_no = len(self.lines)-1
        else:
            line_no = line_no-1

        self.store_state()
        cur = self.get_cursor()
        if col is not None:
            cur.x = col
        cur.y = line_no
        if cur.y >= len(self.lines):
            cur.y = len(self.lines)-1
        self.scroll_to_line(cur.y)
        self.move_cursors()

    def duplicate_line(self):
        """Copy current line and add it below as a new line."""
        curs = sorted(self.cursors, key=lambda c: (c.y, c.x))
        for cursor in curs:
            line = Line(self.lines[cursor.y])
            self.lines.insert(cursor.y+1, line)
            self.move_y_cursors(cursor.y, 1)
        self.move_cursors()
        self.store_action_state("duplicate_line")



from suplemon.linelight.color_map import color_map


class Syntax:
    def get_comment(self):
        return ("//", "")

    def get_color(self, raw_line):
        color = color_map["white"]
        line = raw_line.strip()
        keywords = ("if", "else", "finally", "try", "catch", "foreach",
                    "while", "continue", "pass", "break")
        if line.startswith(("include", "require")):
            color = color_map["blue"]
        elif line.startswith(("class", "public", "private", "function")):
            color = color_map["green"]
        elif line.startswith("def"):
            color = color_map["cyan"]
        elif line.startswith("return"):
            color = color_map["red"]
        elif line.startswith("$"):
            color = color_map["cyan"]
        elif line.startswith(("#", "//", "/*", "*/")):
            color = color_map["magenta"]
        elif line.startswith(keywords):
            color = color_map["yellow"]
        return color



from suplemon.linelight.color_map import color_map


class Syntax:
    def get_comment(self):
        return ("//", "")

    def get_color(self, raw_line):
        color = color_map["white"]
        line = raw_line.strip()
        if line.startswith("function"):
            color = color_map["cyan"]
        elif line.startswith("return"):
            color = color_map["red"]
        elif line.startswith("this."):
            color = color_map["cyan"]
        elif line.startswith(("//", "/*", "*/", "*")):
            color = color_map["magenta"]
        elif line.startswith(("if", "else", "for ", "while ", "continue", "break")):
            color = color_map["yellow"]
        return color
